package PerlMx::Samples::BlackList;

use Net::DNS;
use Socket;
use PerlMx;
use base qw/PerlMx/;
use Config::General;

use strict;

# we want a config file that is terminal name of this package.
my $config_file = __PACKAGE__ . ".conf";
$config_file =~ s/^[\w:]+:://;

# what to do about mail that is ignored for one recipient but accepted
# by another?
# with reject, we can just delete the recipient
# with quarantine, we can delete the recipient and move it to the quarantine
my %actions = map {$_ => 1} ('ignore',	    # don't do anything to this mail
			     'accept',	    # allow the mail but add a spam header
			     'quarantine',  # move the mail to a quarantined area
					    #   (add a spam header for tracking)
			     'reject',	    # reject the mail
			    );

# the following are the legal variables in the message that gets put
# into the X-Spam-Envelope header field
my %msg_vars = map {$_ => 1} ('spammer',    # ip address of the possible spammer
			      'rbl',	    # rbl that flagged this message as spam
			     );
my @boolean_vars = ('Debug',		    # spew out obnoxious messages?
		    'quarantine_keep_body', # keep the body of message when it
					    #    gets put into quarantine.
		    'use_txt_record',       # use the TXT record from the RBL for
					    #    this A req
		   );


my %config = Config::General->new($config_file)->getall();

validate_config(\%config);

sub new {
    bless {
	   config   => \%config,
	   resolver => Net::DNS::Resolver->new,
	   CONNECT  => \&pmx_connect,
	   HELO     => \&pmx_helo,
	   ENVFROM  => \&pmx_envfrom,
	   ENVRCPT  => \&pmx_envrcpt,
	   HEADER   => \&pmx_header,
	   EOH	    => \&pmx_eoh,
	   BODY     => \&pmx_bodychunk,
	   EOM	    => \&pmx_eom,
	   CLOSE    => \&pmx_close,
	}, shift;
}

sub pmx_connect {
    my $ctx = shift;

    $ctx->{recipients}   = [];
    $ctx->{hostname}     = shift;
    my $addr = shift;
    if ($addr) {
	$ctx->{hostaddr}     = inet_ntoa((unpack_sockaddr_in($addr))[1]);
    }
    else {
	warn "Null address for host '$ctx->{hostname}', ignoring\n";
	$ctx->{hostaddr}     = "0.0.0.0";
    }
    $ctx->{full_message} = '$_' . $ctx->{hostname} . " [" . $ctx->{hostaddr} . "]\n";
    $ctx->{full_message} .= "T" . time . "\n";

    return SMFIS_CONTINUE;
}

sub pmx_helo {
    my $ctx = shift;
    $ctx->{full_message} .= "s" . shift() . "\n";
    return SMFIS_CONTINUE;
}

# note that we're throwing away any ESMTP arguments that might've been sent.
sub pmx_envfrom {
    my $ctx = shift;
    $ctx->{full_message} .= "S" . shift() . "\n";
    return SMFIS_CONTINUE;
}

# note that we're throwing away any ESMTP arguments that might've been sent.
sub pmx_envrcpt {
    my $ctx  = shift;
    my $rcpt = shift;

    push @{$ctx->{recipients}}, $rcpt;
    $ctx->{full_message} .= "RPF:$rcpt\n";
			   # It's a (P)rimary (R)ecipient and it has (F)ailed
    return SMFIS_CONTINUE;
}

sub pmx_header {
    my $ctx = shift;
    $ctx->{full_message} .= shift() . ": " . shift() . "\n";
    return SMFIS_CONTINUE;
}

sub pmx_eoh {
    my $ctx = shift;
    $ctx->{full_message} .= "X-Spam-Envelope: \n";
    $ctx->{full_message} .= "\n";  # separate the headers from the body
    return SMFIS_CONTINUE;
}

sub pmx_bodychunk {
    my $ctx = shift;
    $ctx->{config}{quarantine_keep_body} and $ctx->{full_message} .= shift;
    return SMFIS_CONTINUE;
}

sub pmx_eom {
    my $ctx = shift;
    my $res  = $ctx->{resolver};
    my %config = %{$ctx->{config}};
    my $set_quarantine = 0;
    my $set_reject = 0;
    my $spammer = $ctx->{hostaddr};
    my $header_added = 0;

    $ctx->{full_message} .= ".\n";   # terminate the built message

    $spammer =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/;
    if (!defined($4) or $spammer eq "0.0.0.0") {
	warn "sender $spammer didn't match as a valid IP address\n";
	$ctx->addheader('X-Spam-Invalid',
			"sender $spammer didn't match as a valid IP address");
	return SMFIS_CONTINUE;
    }

    my $rev_addr = "$4.$3.$2.$1.";  # note that there *has* to be a dot at the end.

 RBL:
    foreach my $rbl (keys %{$ctx->{config}{RBL}}) {
	next unless keys %{$config{RBL}{$rbl}};
	my $query = $res->query($rev_addr . $rbl, 'A');

	next RBL unless $query;

    ANSWER:
	foreach my $rr ($query->answer) {
	    next unless $rr->type eq 'A';
	    my $spam_message = "";

	    warn "Found $ctx->{hostname} in RBL query" if $config{Debug};
	    if ($config{use_txt_record}) {
		$query = $res->query($rev_addr . $rbl, 'TXT');
		if ($query) {
		    foreach my $rr ($query->answer) {
			next unless $rr->type eq 'TXT';
			$spam_message = $rr->rdatastr;
			last;
		    }
		}
	    }
	    $spam_message ||= $config{spam_message};

	    # sort isn't the best thing to do here but we need to impose
	    # some sort of ordering and we need it to be a hash
	    foreach my $addr (sort keys %{$config{RBL}{$rbl}}) {
		my $action = $config{RBL}{$rbl}{$addr};
		next if $action eq 'ignore';
		my @rcpts;

		# quote the address since it could have funky stuff in it.
		# we're not up to using regexes on the addr (yet?)
		if (@rcpts = grep { /\Q$addr/i } @{$ctx->{recipients}}) {
		    $set_quarantine++ if $action eq 'quarantine';
		    $set_reject++ if $action eq 'reject';
		    if ($action eq 'reject' or $action eq 'quarantine') {
			$ctx->delrcpt($_) foreach @rcpts;
			@{$ctx->{recipients}} = grep { ! /\Q$addr/i } @{$ctx->{recipients}};
			warn "Removing recipient $addr: " . join(", ", @rcpts)
			    if $config{Debug};
		    }

		    if (!$header_added and
			($action eq 'quarantine' or $action eq 'accept'))
		    {
			my $message;
			($message = $spam_message) =~ s/(\$\w+)/$1/gee;
			$ctx->addheader('X-Spam-Envelope', $message);
			$ctx->{full_message} =~ s/^(X-Spam-Envelope: )/$1$message/m;
			$header_added++;
		    }
		}
	    }
	    my $queue_id = $ctx->getsymval('i');  # sendmail macro for queue id; see
						  #    sendmail/dist/doc/op/op.me
	    if (!@{$ctx->{recipients}}) {
		if ($set_quarantine) {
		    warn "Quarantining message '$queue_id'\n" if $config{Debug};
		    # if our write fails, something is wrong with the i/o system,
		    # requeue the message
		    write_quarantine($ctx, $queue_id) || return SMFIS_TEMPFAIL;
		    return SMFIS_REJECT;
		}
		elsif ($set_reject) {
		    warn "Rejecting message '$queue_id\n" if $config{Debug};
		    return SMFIS_REJECT;
		}
		else {
		    warn "How'd I get to no-recipients without quarantine or reject?\n";
		}
	    }
	    elsif ($set_quarantine) {
		# if our write fails, something is wrong with the i/o system,
		# so requeue the message
		write_quarantine($ctx, $queue_id) || return SMFIS_TEMPFAIL;
	    }

	    last ANSWER;
	}
    }

    return SMFIS_CONTINUE;
}

sub pmx_close {
    my $ctx = shift;

    # this could potentially be large, so delete it here rather than on connect
    delete $ctx->{full_message};

    return SMFIS_CONTINUE;
}

sub write_quarantine {
    my ($ctx, $queue_id) = @_;

    my $filename = $ctx->{config}{quarantine_area} . "PMX_" . $queue_id;
    # be real careful about writing these files.
    eval
    {
	open(QUAR, ">$filename") || die "Unable to open $filename for writing: $!";
	print QUAR $ctx->{full_message} || die "Unable to write $filename: $!";
	close(QUAR) || die "Unable to close $filename: $!";
    };

    if ($@) {
	warn $@;
	return;
    }
    return 1;
}

1;

=head1 NAME

PerlMx::Samples::BlackList - Checks for spam hosts

=head1 SYNOPSIS

    pmx1 PerlMx::Samples::BlackList >& BlackList.log &

=head1 DESCRIPTION

PerlMx::Samples::BlackList is a mail filter designed to stop spam on a per-user
basis.

Blacklist works by performing a DNS lookup for a stylized version of the IP address
of the mail host it currently has a connection with.

This filter requires that you have both the Net-DNS and Config-General packages from
CPAN installed.  If you do not have these installed, you can run C<ppm install Net-DNS
Config-General>.

The following documentation is an excerpt from the MAPS documentation on the
subject. The document is available in its entirety at 
http://mail-abuse.org/rbl/usage.html

=over 4

=item *

Direct usage via DNS (subscription may be needed) - The simplest way to get started
using the MAPS RBL to protect your mail relay against theft of service by
spammers is to arrange for it to make a DNS query (of a stylized name) whenever
you receive an incoming mail message from a host whose spam status you do not
know.

The theory of operation is simple. Given a host address in its dotted-quad
form, reverse the octets and check for the existence of an ``A RR'' at that
node under the rbl.maps.vix.com node. So if you get an SMTP session from
[192.5.5.1] you would check for the existence of:

=back

=over 8

=item *

1.5.5.192.rbl.maps.vix.com. IN A 127.0.0.2

=back

=over 4

=item * 

We chose to use an ``A RR'' because that's what Sendmail makes easy to
do. The choice of [127.0.0.2] as the target address was arbitrary but
will not change. As it happens, we supply a bogus MAPS RBL entry for
[127.0.0.2] so that mail transport developers have something to test
against.

If an ``A RR'' is found by this mechanism, then there will also be a
``TXT RR'' at the same DNS node. The text of this record will be
suitable for use as a reason text for a bounced mail notification.
Currently the text is constant and currently there is no way to use it
from Sendmail, but there it is anyway.>

=back

=head1 CONFIGURATION

The B<%lists> hash at the top of the filter is keyed on the domain that 
provides the service.  The values are lists of regular expressions that 
match the addresses of people who do not wish to receive SPAM from 
hosts found in the corresponding list. The following lists are 
run by MAPS:

	blackholes.mail-abuse.org
	relays.mail-abuse.org
	dialups.mail-abuse.org

=head1 DEPENDENCIES

* Net::DNS

* Config::General

=head1 COPYRIGHT

Copyright (c) 2000-2003 ActiveState.  All rights reserved.

=head1 SEE ALSO

L<perl>, L<PerlMx>, L<Net::DNS>, L<Config::General>

=cut

#' sop for X?emacs parsing

# I'm using this like so: perl -MBlackList -e BL_dump_config
sub dump_config {
    require Data::Dumper;

    print Data::Dumper->Dumpxs([\%config], ["config"]);
}

*main::BL_dump_config = *PerlMx::Samples::BlackList::dump_config if $config{Debug};

sub validate_config {
    my $config = shift;

    $config->{default_action}  ||= 'quarantine';
    $config->{quarantine_area} ||= '/var/spool/PerlMx/quarantine/';
    $config->{spam_message}    ||= 'Mail from $spammer refused by blackhole site $rbl';
    $config->{use_txt_record}  ||= 'false';

    check_action('default_action', $config->{default_action});

    # make sure the directory ends in a /
    $config->{quarantine_area} .= "/" unless substr($config->{quarantine_area}, -1) eq '/';

    for my $rbl (keys %{$config->{RBL}}) {
	foreach my $rcpt (keys %{$config->{RBL}{$rbl}}) {
	    $config->{RBL}{$rbl}{$rcpt} =~ tr [A-Z] [a-z];
	    check_action($rcpt, $config->{RBL}{$rbl}{$rcpt});
	    $config->{RBL}{$rbl}{$rcpt} ||= $config->{default_action};
	}
    }

    foreach my $var ($config->{spam_message} =~ /\$(\w+)/g) {
	$msg_vars{$var} or die "Invalid message variable: '$var' in '" . $config->{spam_message} . "'\n";
    }

    # fix the values of the boolean variables in the config file.  Now they
    # conform to the Perl way.
    foreach my $var (@boolean_vars) {
	# alias the variable to $_ to make our regexes look clean
	local *_ = \$config->{$var};
	/^no$/i    and $_ = 0;
	/^false$/i and $_ = 0;
	/^off$/i   and $_ = 0;
    }
}

sub check_action {
    my ($key, $action) = @_;

    # it's okay for the action to not be defined, we use the default then
    return unless defined($action);

    die "Invalid action '$action' for '$key'\n" unless $actions{$action};
}

1;
