###############################################################################
#
# PerlMx::Samples::DeMultiplexor
#
# A filter that will add one or more recipients to a messages recipient
# list based on addresses in the current recipient list.
#
# Michael Smith <mikesm@ActiveState.com>
#
# Copyright (c) 2000-2003 ActiveState. All rights reserved.
#
###############################################################################
package PerlMx::Samples::DeMultiplexor;

use PerlMx;
use base qw/PerlMx/;

# Configure the %recipients hash as required
my %recipients = (
	'mikesm@ActiveState\.com' => 'stuntman@ActiveState.com',
    );

sub new {
    my $class = shift;
    bless {
		HEADER  => \&cb_header,
		EOM	=> \&cb_eom
    }, $class;
}

sub cb_header {
    my $self = shift;
    my $name = shift;
    my $value = shift;
    my $pattern;
    my $addrs;

    if($name =~ /^(to|cc)$/i) {
        while(($pattern, $addrs) = each(%recipients)) {
	    if($value =~ /$pattern/im) {
		push @{$self->{new_recipients}}, split /,/, $addrs;
	    }
	}
    }

    return SMFIS_CONTINUE;
}

sub cb_eom {
    my $self = shift;
    if(exists $self->{new_recipients}) {
	foreach my $rcpt (@{$self->{new_recipients}}) {
	    $self->addrcpt($rcpt);
	}
	delete $self->{new_recipients};
    }
    return SMFIS_ACCEPT;
}

1;

__END__

=head1 NAME

PerlMx::Samples::DeMultiplexor -- Adds addresses to messages

=head1 SYNOPSIS

    pmx1 PerlMx::Samples::DeMultiplexor >& DeMultiplexor.log &

=head1 DESCRIPTION

A filter that adds addresses to the recipient list of a message based
on the addresses encountered in the To: and CC: header fields.

=head1 CONFIGURATION

All configuration is managed by adding keys/values to the B<%recipients> hash
in the code at the top of the filter. The keys in this hash are regular 
expressions that may match addresses in the To: or CC: header fields of the
message. The values in this hash are a comma delimited list of addresses to 
be added to the recipient list if the corresponding regular expression locates
a match.

For example:

    my %recipients = (
	'mikesm@ActiveState\.com' => 'stuntman@ActiveState.com',
    );

In this example, when the filter finds the email address 'mikesm@ActiveState.com'
in the To: or CC: fields of the message header, the filter adds 
'stuntman@ActiveState.com' added to the message recipient list.

B<NOTE>: Keys in the %recipients hash are regular expressions used to match
addresses in the To: and CC: fields; therefore, you may need to escape
special pattern matching characters you intend to match the literal character.

B<NOTE>: This filter does not match addresses found in the envelope of the
message.

=head1 COPYRIGHT

Copyright (c) 2000-2003 ActiveState.  All rights reserved.

=head1 SEE ALSO

L<perl>, L<PerlMx>

=cut
