###############################################################################
#
# PerlMx::Samples::Evil_Filter
#
# Inspired by John Hardin's procmail scripts this filter is designed to 
# reattach nasty content after archiving it in tar.gz format.
#
# Will Whittaker <willw@ActiveState.com>
#
# Copyright (c) 2000-2003 ActiveState. All rights reserved.
#
###############################################################################
package PerlMx::Samples::Evil_Filter;

use strict;
use PerlMx::MIME_Filter;
use Compress::Zlib;
use base qw /Exporter PerlMx::MIME_Filter/;

# mail admin's address, you should probably change this
my $admin='postmaster@localhost';

# list of regexes to match known trojans
# anything matching this list will just be rejected
my @evil = (
    '.+\.chm',
    '.+\.hlp',
    '.+\.hta',
    '.+\.js',
    '.+\.shb',
    '.+\.shs',
    '.+\.vb',
    '.+\.vbe',
    '.+\.vbs',
    '.+\.wsf',
    '.+\.wsh',
    'IBMls\.exe',
    'aol4free\.com',
    'babylonia\.exe',
    'badass\.exe',
    'buhh\.exe',
    'chocolate\.exe',
    'compu_ma\.exe',
    'happy99\.exe',
    'i-watch-u\.exe',
    'ie0199\.exe',
    'jesus\.exe',
    'list\.doc',
    'path\.xls',
    'photos17\.exe',
    'picture\.exe',
    'pretty park\.exe',
    'prettypark\.exe',
    'serialz\.hlp',
    'setup\.exe',
    'story\.doc',
    'suppl\.doc',
    'surprise\!\.exe',
    'x-mas\.exe',
    'y2kcount\.exe',
    'yahoo\.exe',
    'zipped_files\.exe'
);

# anything matching this list will be gzipped and reattached
# with a warning
my @dangerous = (
    '\.html?$',
    '\.exe$',
    '\.com$',
    '\.cmd$',
    '\.bat$',
    '\.pif$',
    '\.sc[rt]$',
    '\.lnk$',
    '\.do[ct]$',
    '\.xl[sw]$',
    '\.p[po]t$',
    '\.rtf$',
    '\.vb[se]?$',
    '\.hta$',
    '\.pm$',
    '\.plx?$',
    '\.pyc?$',
    '\.sh[bs]$',
    '\.hlp$',
    '\.chm$',
    '\.eml$',
    '\.ws[cfh]$',
    '\.ad[ep]$',
    '\.jse?$',
    '\.md[be]$',
    '\.ms[ip]$',
    '\.reg$',
    '\.asd$'
);

sub new {
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->register_mimetype('raw',\&cb_raw);
    return $self;
}

sub cb_raw {
    my $ctx = shift;
    my $e = shift;
    my $ne = $e->dup;
    my $warning = "";

    # these are considered too dangerous to keep sending they will be rejected
    # immediately. add new stuff here as necessary

    my $hdchck = headercheck($e->stringify_header);
    if ($hdchck) {
        warn localtime() . " RFC 822 Header exceeded 256 characters:\n"
	     . "$hdchck\nMail REJECTED\n";
        return REJECT;
    }
    my $mmchck = mimeheadercheck($e);
    if ($mmchck) {
        warn localtime() . " mime header exceeded 128 characters:\n"
	     . "$mmchck\nMail REJECTED\n";
        return REJECT;
    }
    my $bad = evilfilecheck($e,\@evil);
    if (scalar keys %$bad) {
        warn localtime() . " Potential trojan attached to message\n"
	     . join(" ", values %$bad)
	     . "\nMail REJECTED\n";
        return REJECT;
    }
  
    # this will get to the recipient but will gzip dangerous attachments and
    # add a warning. add new stuff here as necessary (a macro scanner for
    # office attachments and an active html scanner would both be nice)

    $bad = evilfilecheck($e,\@dangerous);
    if (scalar keys %$bad) {
        warn localtime() . " Potential trojan attached to message.\n"
	     . "Original email is attached as gzipped file\n";
        # send warning about evil files to admin, quarantine file in gz
	# continue sending
        $warning .= "Potential trojan attached to message.\n"
	          . "Original file is attached as gzipped file\n"
		  . "\n\nSending executable content in mail is dangerous, "
		  . "please ask for a link next time\n\n";

        # this gets around the ugliness of MIME::Entity's halfbaked ability
	# to manipulate attachments

        my $i = 0;
        my @keep;
        my @kill;
        my @killname;
        foreach my $part ($e->parts) {
            if ($$bad{$i}) {
                push @kill, $part;
                push @killname, $$bad{$i};
            }
            else {
                push @keep, $part;
            }
            $i++;
        }

        # now that we know what we want, we can keep the good stuff
        $e->parts(\@keep);

        # attach a warning about the bad stuff
        $e->attach(Type     => 'text/plain',
                   Encoding => '7bit',
                   Data     => "$warning\n");

        # and attach all the dangerous stuff as gzipped files

        for ($i = 0; $i < @kill; $i++) {
            my $kill = $kill[$i];
            my $filename = $killname[$i];
            $filename =~ s/\./\-/g;
            $filename .= int(rand(9)).int(rand(9)).int(rand(9)).".gz";
            my $compressed = Compress::Zlib::memGzip($kill->stringify);  
            $e->attach(Type     => 'application/octet-stream',
                       Encoding => 'base64',
                       Filename => $filename,
                       Data     => $compressed);
        }

        # cc the admin here, comment this out if you don't want the admin cc'ed
        # on every potentially dangerous email
        $ctx->addrcpt($admin);
    }
    return CONTINUE;
}


# checks if RFC 822 headers are > 256 chars
sub headercheck {
    my $head = shift;
    my @head;
    @head = split /\n/, $head;
    foreach my $line (@head) {
        if (length($line) > 256) {
            return $line;
        }
    }
    return undef;
}

# checks if mime header is greater than 128 chars
sub mimeheadercheck {
    my $e = shift;
    my @parts;
    if (my $bh = $e->bodyhandle) {
        my $head = $e->head;
        my @head = split "\n", $head;
        foreach my $header (@head) {
            if (length($header) > 128) {
                return $header;
            }
        }
    }
    else {
        @parts = $e->parts;
        foreach my $part (@parts) {
            my $head = $part->head;
            my @head = split "\n", $head;
            foreach my $header (@head) {
                if (length($header) > 128) {
                    return $header;
                }
            }
        }
    }
    return undef;
}

# checks if filename matches known trojans, part place and filename as a
# hash ref
sub evilfilecheck {
    my $e = shift;
    my $evil = shift;
    my @parts;
    my %bad;
    my $head;
    if (my $bh = $e->bodyhandle) {
        $head = $e->head;
        if (evilpath($head->recommended_filename,$evil)) {
            $bad{0} = $head->recommended_filename;
            return \%bad;
        }
    }
    else {
        @parts = $e->parts;
        my $i = 0;
        foreach my $part (@parts) {
            $head = $part->head;
            if (evilpath($head->recommended_filename,$evil)) {
                $bad{$i} = $head->recommended_filename;
            }
            $i++;
        }
    }
    return \%bad;
}

# checks paths against trojan list
sub evilpath {
    my $path = shift;
    my $evil = shift;
    foreach my $trojan (@$evil) {
        if ($path =~ /$trojan/i) {
            return $trojan;
        }
    }
    return undef;
}

1;


__END__

=head1 NAME

PerlMx::Samples::Evil_Filter - Removes malware from email

=head1 SYNOPSIS

    pmx1 PerlMx::Samples::Evil_Filter >& Evil_Filter.log &

=head1 DESCRIPTION

This filter rejects content with oversized RFC 822 headers, oversized
MIME headers, and known trojans. This filter also removes a list of potentially
dangerous file extensions and re-attaches them in gzipped form with a warning.

You need to change the $admin variable to point to the appropriate role account.
You also need to edit @evil and @dangerous as appropriate for your security
policy.  File names matched by @evil will be rejected at the gateway, so BE
CAREFUL with the contents @evil.  File names matched by @dangerous will be delivered,
but they will be gzipped. This will break any additional virus checking
software you might have, so BE CAREFUL.  One more warning: the RFC 822 and MIME
header checkers are not based on any RFC, but on a "safe" value instead.  Legitimate
mail can trip this.  Monitor your logs; if it happens more often than acceptable,
adjust the number of chars in the check function appropriately.

=head1 TODO

* Add MS Office macro scanning and active HTML scanning.  

* Add support for various virus scanners.  

* Use a conf file instead of putting the evil content in the .pm.  

* Better info to STDERR would probably be helpful.  

=head1 DEPENDENCIES

* Compress::Zlib

=head1 COPYRIGHT

Copyright (c) 2000-2003 ActiveState. All rights reserved.


=cut

