##############################################################################
#
# PerlMx::Samples::External
#
# General purpose filter that can do a number of things based on
# a simple configuration file. You can configure this module to do any of
# the following:
#
#   o add headers
#   o add a signature
#   o execute arbitrary commands that will filter a document inplace on
#     the file system
#
# Michael Smith <mikesm@ActiveState.com>
#
# Copyright (c) 2000-2003 ActiveState. All rights reserved.
#
##############################################################################
package PerlMx::Samples::External;

use strict;
use File::Temp;
use PerlMx::MIME_Filter;
use base qw ( PerlMx::MIME_Filter );

our $DEBUG = 0;
our $config_file = shift || 'PerlMx-External.conf';
our %config = parse_config();

#
# Constructor
#
sub new {
    my $class = shift;
    my $self = $class->SUPER::new();
    
    # Register our callback
    $self->register_mimetype('raw', \&cb_raw);

    # Construct and set handler for each known
    # MIME-Type
    foreach my $mime_type (keys %{$config{Commands}}) {
	$self->register_mimetype($mime_type, construct_cmd_callback($mime_type));
    }

    foreach my $mime_type (keys %{$config{Replace}}) {
	$self->register_mimetype($mime_type, construct_replace_callback($mime_type));
    }
    
    return $self;
}

#
# Our callback
#
sub cb_raw {
    my $ctx = shift;
    my $msg = shift;

    # Add headers
    addheaders($ctx);

    # Sign message
    sign($msg);

    # Return; ok
    return CONTINUE;
}

#
# Runs individual command
#
sub exec_cmd {
    my $cmd = shift;
    my $filename = shift;
    my $h = IO::File->new;
    my $cmdline = sprintf("$cmd", $filename);
    my $status = 0;


    if($h->open("$cmdline |")) {
	$status = $? >> 8;
	if($DEBUG) {
	    warn "\n\n", '=' x70, "\n\n";
	    warn <$h>;	
	    warn "\n\n", '=' x70, "\n\n";
	}
	$h->close;
    }
    else {
	warn "Could not execute command: '$cmdline'";
    }
    
    return $status;
}

#
# Add headers to the message
#
sub addheaders {
    my $msg = shift;
    foreach my $header (@{$config{Headers}}) {
	$msg->addheader(split(/:\s*/, $header));
    }
}

#
# Sign the message
#
sub sign {
    my $msg = shift;
    if( map { $msg->head->get('From') =~ /$_/i } @{$config{Domains}} ) {
	$msg->sign(Attach => 1, Signature => $config{Signature})
	    if length($config{Signature}) > 0;
    }
}

#
# Remove the MIME::Entity referenced by $e. $e must be decendent from $top
#
sub remove_part {
    my $top = shift;
    my $e   = shift;
    $top->parts( [ grep {!/\Q$e\E/} $top->parts ] );
}

#
# Add a simple notification message to the provided MIME::Entity object.
# In the future this should call a function in PerlMx::Utils
#
sub notify {
    my $e = shift;
    $e->attach(Data => sprintf(shift, @_),
	       Type => 'text/plain');
}

#
# Construct a callback for external commands
#
sub construct_cmd_callback {

    my $mime_type = shift;
    
    # Note the handler subroutine uses closure to 
    # remember the command to be executed
    return sub {
	my $ctx = shift;
	my $msg = shift;
	my $command = $config{Commands}{$mime_type};
	my ($tmpfh, $file) = File::Temp::tempfile();
	my $fh = $msg->open('r');
	my $status = 0;

	if ($fh) {
	    while($_ = $fh->getline) {
		print $tmpfh $_;
	    }
	    $fh->close();
	    $status = exec_cmd( $command, $file );
	    
	    if ( -e $file ) {
		$fh = $msg->open('w');
		if($fh) {
		    seek($tmpfh, 0, 0);
		    $fh->print(<$tmpfh>);
		    $fh->close();
		}
	    }
	    else {
		remove_part($ctx->{__perlmxm_me}, $msg);
		notify( $ctx->{__perlmxm_me}, 
			"Attachment %s, from %s was removed", 
			$msg->head->recommended_filename, 
			$msg->head->get('From:') );
	    }
	
	}

	close $tmpfh if $tmpfh;
	unlink $file if $file;

	return CONTINUE;
    };
}

#
# Construct a callback to do search and replace
#
sub construct_replace_callback {
    my $mime_type = shift;

    # Note the handler subroutine uses closure to 
    # remember the search and replace patterns
    return sub {
	my $ctx = shift;
	my $msg = shift;
	my @search_rep = @{$config{Replace}{$mime_type}};
	my @lines;
	my $line;
	my $fh = $msg->open('r');
	
	if($fh) {
	    while($line = $fh->getline) {
		#map { eval "\$line =~ $_" } @search_rep;
		
		map { 
		    my ($j1, $pat, $j2, $rep) = 
			/^('?)(.+[^\\])\1\s+('?)(.+[^\\])\3$/; 
		    $line =~ s/$pat/$rep/ig;
		} @search_rep;
		
		push @lines, $line;
	    }
	    $fh->close();
	    $fh = $msg->open('w');
	    if($fh) {
		$fh->print(@lines);
		$fh->close();
	    }
	}
	
	return CONTINUE;
    };
}

#
# Parse our simple configuration file
#
sub parse_config {
    my %config;
    my $in_sig = 0;
    my $fh = IO::File->new;

    # Open config file
    $fh->open($config_file) or warn "$config_file : $!";
    
    while(<$fh>) {
	
	chomp unless $in_sig;

	if( /^<signature\s*>\s*$/i ) {
	    $in_sig = 1;
	}
	elsif( /^<\/signature\s*>\s*$/i ) {
	    $in_sig = 0;
	}
	elsif( $in_sig ) {
	    $config{Signature} .= $_; 
	}
	elsif(/^#|^\s*$/) {  # This is for comments
	    next;
	}
	elsif( /^Domain\s+(.+?)\s*$/ ) {
	    push @{$config{Domains}}, $1;
	}
	elsif( /^Header\s+(.+?)\s*$/ ) {
	    push @{$config{Headers}}, $1;
	}
	elsif( /^Command\s+(.+?)\s+(.+?)\s*$/ ) {
	    $config{Commands}{$1} = $2;
	}
	elsif( /^Replace\s+(.+?)\s+(.+?)\s*$/ ) {
	    push @{$config{Replace}{$1}}, $2;
	}
	else {
	    warn "Unknown configuration option '$_'";
	}

    }

    # Close configuration file and return 
    # configuration info
    $fh->close if keys %config;
    return %config;
}

1;

__END__

=head1 NAME

PerlMx::Samples::External -- Performs tasks based on a configuration file

=head1 SYNOPSIS

    pmx1 PerlMx::Samples::External -- config.txt >& External.log &

=head1 DESCRIPTION

General purpose filter that can do a number of things based on
a simple configuration file. The configuration file is included after the 
'--' on the command line. You can configure this module to do any of
the following:

=over 4

=item *

add headers

=item *

add a signature

=item *

execute arbitrary commands that will filter a document in place on the file
system

=item *

search for and replace text found in attachments of a specified MIME type

=back

=head1 CONFIGURATION

This filter uses a simple text based configuration file. Lines beginning with
`#' are considered comments and are ignored by the filter unless they occur
within a configuration block such as the E<lt>signatureE<gt> ... E<lt>/signatureE<gt>
block described below.

=head2 Domain

Used to specify domains that are considered local. Certain functions of the 
filter are enabled or disabled based on whether or not the remote host is
considered part of the local mail domain. For instance, signatures are only
added to message originating in the local domain. Following is a sample
portion of the configuration file that specifies three local domains. 

    #
    # Domains that are considered local
    #
    Domain activestate.com
    Domain perlclinic.com
    Domain mdms.org

=head2 Signature

This section of the configuration file specifies a signature that should be
attached to all outgoing messages originating from any of the local domains.
Following is a sample portion of the configuration file that specifies a 
signature to be added to outgoing messages.

    #
    # Include a signature in all messages
    #
    <signature>
    ###############################################################
    # 
    # This message was filtered by ActiveState PureMessage.
    #
    # PureMessage is ActiveState's solution for rapidly developing
    # and integrating, fast and flexible email based content
    # management.
    #
    #                                        You need PureMessage!
    #
    ###############################################################
    </signature>

=head2 Header

The B<Header> configuration option specifies a header field that is to be 
added to outgoing messages. The header name should be separated from the header
value by a `:'. Following is a sample portion of the configuration file that
specifies two header fields to be added to outgoing messages.

    #
    # Add these headers to the message
    #
    Header X-FilteredBy: ActiveState PureMessage
    Header X-Comment: You need PureMessage!

=head2 Command

The B<Command> configuration option specifies an external program to be
executed to filter attachments of a specific MIME type. Fields in this 
command are separated by spaces. The first field specifies the MIME type of
the attachments we are interested in filtering. The remaining portion of the
configuration option is the command line of the external program to be run.
The first occurrence of `%s' in the command line portion of the option will 
be replaced by the name of the file to be filtered in place on the file system,
therefore you will need to include `%s' once in your command line. Following
is a sample portion of the configuration file that specifies an external 
program that will filter all attachments with the MIME type 
`application/msword'.

    #
    # Execute these commands
    #
    Command application/msword cp -v %s /home/mikesm/tmp

=head2 Replace

The B<Replace> configuration option specifies that all occurrences of the
string in field two of the option be replaced by the string in field three of
the option for all attachments of the the MIME type specified in field one
of the option. Following is a sample portion of the configuration file that
specifies two replacements to be done for all attachments with the MIME type
`text/html':

    #
    # Search and replace
    #
    Replace text/html <script <script_defanged
    Replace text/html </script </script_defanged

=head1 DEPENDENCIES

* File::Temp

=head1 COPYRIGHT

Copyright (c) 2000-2003 ActiveState. All rights reserved.

=head1 SEE ALSO

L<PerlMx>, L<MIME::Entity>, L<File::Temp>, L<perl>

=cut
