#
# A sample to illustrate a Filter in action.
#
# This is the filter implementation.  See also pmx1 (the filter
# driver).
#

package PerlMx::Samples::GroupedWordFilter;

use 5.006;
use strict;
use warnings;

use PerlMx;
use base qw(PerlMx);

use Sys::Hostname;
use Fcntl;

sub LOCK_SH() { 1 }
sub LOCK_EX() { 2 }
sub LOCK_UN() { 8 }

# variables
my $config_file="GroupedWordFilter.conf";    # name of the config file and word list
my %badwords;

print "INIT()\n";

open INIT, $config_file;
flock (INIT, LOCK_SH);
my @t = <INIT>;
flock (INIT, LOCK_UN);
close INIT;

my $in_block = 0;
my $priority = 0;
my $ptr = 0;
my $block = "";
foreach (@t) {
  if (!$in_block) {
    # if we're not in a block and we see the start-of-block:
    if (/^(sub\s*\{.*)/) {
      
      # if this is the first sub, eval the preamble
      if ($ptr == 0) {
	#print "evaluating preamble:\n$block\n";
	eval($block);
	warn $@ if $@;
      }
      
      $in_block = 1;
      $block = $1;
    }
    # if the line is a comment or blank, continue;
    elsif (/^\s*\#/ || /^\s*$/) {
      next;
    }
    # if the line is before the first subroutine, build the $block up
    # to prepare to eval()
    elsif ($ptr == 0) {
      $block = $block . $_;
    }
    # if we're not in a block, and we see a non-start-of-block,
    # then we should add a bad-word to the list of names:
    else {
      chomp;
      #print "adding badword: $_\n";
      $badwords{$_} = [ $ptr, $priority ];
    }
  }
  else {
    # if we're in a block and we get an endblock, end the block
    if (/^\}/) {
      $in_block = 0;
      $block = "\$ptr = " . $block . "}";
      #print "evaluating block:\n$block\n";
      eval($block);
      warn $@ if $@;
      $block = "";
      $priority++;
    }
    elsif (/^\s*\#/ || /^\s*$/) {
      next;
    }
    # if we're in a block and we get a non-endblock, push onto block
    else {
      $block = $block . $_;
    }
  }
}

# utility functions

sub cleanup {
    my $ctx = shift;
    my $ok = shift;
    my $status = SMFIS_CONTINUE;

    return $status;
}

# callbacks

sub cb_connect {
    my $ctx = shift;

    # reset things to default values
    $ctx->{func_ptr} = 0;
    $ctx->{words_found} = {};
    $ctx->{sender} = "";
    $ctx->{rcpts} = [];
    $ctx->{headers} = [];
    $ctx->{cbody} = "";
    $ctx->{priority} = 0;

    return SMFIS_CONTINUE;
}

sub cb_envfrom {
    my $ctx = shift;

    $ctx->{sender} = shift;

    return SMFIS_CONTINUE;
}

sub cb_envrcpt {
    my $ctx = shift;
    my $rcpt = shift;

    # add the latest recipient
    push (@{$ctx->{rcpts}}, $rcpt);

    return SMFIS_CONTINUE;
}

sub cb_header {
    my $ctx = shift;
    my $headerf = shift;
    my $headerv = shift;

    push (@{$ctx->{headers}}, [ $headerf, $headerv ]);

    return SMFIS_CONTINUE;
}

sub cb_body {
    my $ctx = shift;
    my $body = shift;

    $ctx->{cbody} = $ctx->{cbody} . $body;

    print "Body chunk=\n'$body'\n";

    foreach (keys %badwords) {
	print "checking '$_'...\n";
	if ($body =~ /$_/i) {
	    print "bad word found: '$_'\n";
	    print "highest priority: $ctx->{priority}\n";
	    print "new priority: $badwords{$_}[1]\n";
	    $ctx->{words_found}{$_}++;
	    if ($ctx->{priority} == 0 ||
		$ctx->{priority} > $badwords{$_}[1])
	    {
		print "setting function to call to new priority\n";
		$ctx->{'func_ptr'} = $badwords{$_}[0];
		$ctx->{'priority'} = $badwords{$_}[1];
	    }
	    else {
		print "still calling the higher priority function\n";
	    }
	}
    }

    return SMFIS_CONTINUE;
}

sub cb_eom {
    my $ctx = shift;
    my $result = SMFIS_ACCEPT;

    if ($ctx->{func_ptr} != 0) {
	$result = $ctx->{func_ptr}->( $ctx );
    }
    else {
	$result = $ctx->cleanup(1);
    }

    return $result;
}

sub cb_abort {
    my $ctx = shift;
    return $ctx->cleanup(0);
}

sub cb_close {
    return SMFIS_ACCEPT;
}

sub new {
    my $class = shift;
    my $s = {
		NAME	=> "GroupedWordFilter",
	        CONNECT => \&cb_connect,
		ENVFROM	=> \&cb_envfrom,
		ENVRCPT	=> \&cb_envrcpt,
		HEADER	=> \&cb_header,
		BODY	=> \&cb_body,
		EOM	=> \&cb_eom,
		ABORT	=> \&cb_abort,
		CLOSE	=> \&cb_close,
	    };
    return bless $s, $class;
}

warn "Loading " . __PACKAGE__ . "\n";

1;

__END__

=head1 NAME

PerlMx::Samples::GroupedWordFilter - Scans message body for
keywords

=head1 SYNOPSIS

    pmx1 PerlMx::Samples::GroupedWordFilter >& GroupedWordFilter.log &

=head1 DESCRIPTION

A simple filter that scans the body for key words which are
read from a configuration file on startup.  The configuration file
contains instructions for what should be done if a key word is found.

=head1 HIGH-LEVEL OVERVIEW

This filter is designed to scan the body of emails for key words. The
user provides a custom function for each group of words; the function
is called at the end of the message. Because only one function can be
called, the word groups are assigned a priority starting from 1 and
increasing from there. The first word group defined in the
configuration is priority 1, the second is 2, and so on. The highest
priority is 1 (i.e., the first group of words defined in the
configuration file are the highest priority, and priority decreases
with more groups). If more than one word from a group is found, the
function is still only called once.

Note that in this filter, the entire message is stored in memory until
the end of the message, and no checking for extremely large messages
is performed.

=head1 DETAIL

This section gives as much detail as possible about the code.

=head2 Preamble

  package GroupedWordFilter;

This line lets Perl know that this file provides the package
GroupedWordFilter.

  use 5.006;
  use strict;
  use warnings;

This chunk gives Perl instructions about how to compile the code. For
example, use 5.006 means that this module must be compiled with Perl
5.006 or above. use strict and use warnings control how picky the
compiler is about warning for things like undeclared variables, and so
on.

  use PerlMx;
  use base qw(PerlMx);

This imports the base class into the filter.

  sub LOCK_SH() { 1 }
  sub LOCK_EX() { 2 }
  sub LOCK_UN() { 8 }

These lines define new subroutines which simply return the numbers 1,
2, and 8 respectively. They are constants used in the flock() command.

  my $config_file="GroupedWordFilter.conf";
  my %badwords;

This chunk defines two global variables, $config_file and
%badwords. The first is the filename of the configuration file, which
contains the list of "bad words" and a function which is called if the
bad word is found. The second, %badwords, is the hash in which the words
in the configuration file are stored which the filter is running.

  open INIT, $config_file;
  flock (INIT, LOCK_SH);
  my @t = <INIT>;
  flock (INIT, LOCK_UN);
  close INIT;

This chunk of code reads the configuration file into an array. Each line
of the file is one element of the array.

  my $in_block = 0;
  my $priority = 0;
  my $ptr = 0;
  my $block = "";
  foreach (@t) {
    if (!$in_block) {
      # if we're not in a block and we see the start-of-block:
      if (/^(sub\s*\{.*)/) {
      
        # if this is the first sub, eval the preamble
        if ($ptr == 0) {
	  #print "evaluating preamble:\n$block\n";
	  eval($block);
	  warn $@ if $@;
        }
      
        $in_block = 1;
        $block = $1;
      }
      # if the line is a comment or blank, continue;
      elsif (/^\s*\#/ || /^\s*$/) {
        next;
      }
      # if the line is before the first subroutine, build the $block up
      # to prepare to eval()
      elsif ($ptr == 0) {
        $block = $block . $_;
      }
      # if we're not in a block, and we see a non-start-of-block,
      # then we should add a bad-word to the list of names:
      else {
        chomp;
        #print "adding badword: $_\n";
        $badwords{$_} = [ $ptr, $priority ];
      }
    }
    else {
      # if we're in a block and we get an endblock, end the block
      if (/^\}/) {
        $in_block = 0;
        $block = "\$ptr = " . $block . "}";
        #print "evaluating block:\n$block\n";
        eval($block);
        warn $@ if $@;
        $block = "";
        $priority++;
      }
      elsif (/^\s*\#/ || /^\s*$/) {
        next;
      }
      # if we're in a block and we get a non-endblock, push onto block
      else {
        $block = $block . $_;
      }
    } 
  }

This chunk of code parses the configuration file into the preamble, function
blocks, and individual words. It is a simple state-machine based parser. The
preamble and functions are passed to eval(), which compiles them and runs
them. The words are stored in the hash %badwords.

  sub cleanup {
    my $ctx = shift;
    my $ok = shift;
    my $status = SMFIS_CONTINUE;

    return $status;
  }

This sub, cleanup(), does nothing in this version, but could theoretically 
be used to do any final cleanup on the filter before the message was sent.

=head2 Callbacks

  sub cb_connect {
    my $ctx = shift;

    # reset things to default values
    $ctx->{func_ptr} = 0;
    $ctx->{words_found} = {};
    $ctx->{sender} = "";
    $ctx->{rcpts} = [];
    $ctx->{headers} = [];
    $ctx->{cbody} = "";
    $ctx->{priority} = 0;

    return SMFIS_CONTINUE;
  }

This sub is called when sendmail receives an incoming SMTP connection. This 
function is used to initialize extra values in the object, which are used to
store temporary information about the message. At the end of the message, these
values are examined in order to determine what to do.

  sub cb_envfrom {
    my $ctx = shift;

    $ctx->{sender} = shift;

    return SMFIS_CONTINUE;
  }

This sub is called when sendmail receives an envelope from message. It
contains the address of the sender as an argument. It is used to store
the address of the sender in the object.

  sub cb_envrcpt {
    my $ctx = shift;
    my $rcpt = shift;

    # add the latest recipient
    push (@{$ctx->{rcpts}}, $rcpt);

    return SMFIS_CONTINUE;
  }

This sub is called once for each recipient of the message. It is used to
store the recipient in the object for later use.

  sub cb_header {
    my $ctx = shift;
    my $headerf = shift;
    my $headerv = shift;

    push (@{$ctx->{headers}}, [ $headerf, $headerv ]);

    return SMFIS_CONTINUE;
  }

This sub is called once for each header in the message. It is used to
store the headers for possible use later.

  sub cb_body {
    my $ctx = shift;
    my $body = shift;

    $ctx->{cbody} = $ctx->{cbody} . $body;

    print "Body chunk=\n'$body'\n";

    foreach (keys %badwords) {
	print "checking '$_'...\n";
	if ($body =~ /$_/i) {
	    print "bad word found: '$_'\n";
	    print "highest priority: $ctx->{priority}\n";
	    print "new priority: $badwords{$_}[1]\n";
	    $ctx->{words_found}{$_}++;
	    if ($ctx->{priority} == 0 ||
		$ctx->{priority} > $badwords{$_}[1])
	    {
		print "setting function to call to new priority\n";
		$ctx->{'func_ptr'} = $badwords{$_}[0];
		$ctx->{'priority'} = $badwords{$_}[1];
	    }
	    else {
		print "still calling the higher priority function\n";
	    }
	}
    }

    return SMFIS_CONTINUE;
  }

This function is called for each chunk of the body. It searches
through the body for each item in %badwords (each bad word). If it
finds a bad word it saves that information in the object. The
highest-priority found word is the one whose handler is saved for
later calling.

  sub cb_eom {
    my $ctx = shift;
    my $result = SMFIS_ACCEPT;

    if ($ctx->{func_ptr} != 0) {
	$result = $ctx->{func_ptr}->( $ctx );
    }
    else {
	$result = $ctx->cleanup(1);
    }

    return $result;
  }

This function is called at the end of the message, if the abort
function is not called. Its purpose is to call the user-provided
handler of the highest priority word found, if any.

sub cb_abort {
    my $ctx = shift;
    return $ctx->cleanup(0);
}

This function is called if sendmail detected an external abort
condition, such as a premature disconnection.

sub cb_close {
    return SMFIS_ACCEPT;
}

This function is called when an SMTP connection is closed.

=cut
