#
# A sample to illustrate a filter in action.
#
# This is the filter implementation.  See also pmx1 (the filter
# driver).
#

package PerlMx::Samples::SampleFilter;

use 5.006;
use strict;
use warnings;

use PerlMx;
use base qw(PerlMx);

use Sys::Hostname;
use Fcntl;

# utility functions

sub make_temp_file {
    my $seq = int rand(10000000);
    my $file = "/tmp/smfi$seq";
    my $fh;
    my $iter = 0;
    my $flags = O_CREAT|O_EXCL|O_RDWR|O_BINARY;
    # XXX Real code should use File::Temp::mkstemp() but that's not
    # available in Perl v5.6.0.
    while (!sysopen($fh, $file, $flags, 0600)) {
	$seq = int rand(10000000);
	$file = "/tmp/smfi$seq";
	return if ++$iter > 10;
    }
    return ($file,$fh);
}

sub cleanup {
    my $ctx = shift;
    my $ok = shift;
    my $status = SMFIS_CONTINUE;
    if (exists $ctx->{fh} and !close($ctx->{fh})) {
	unlink($ctx->{fname});
	$status = SMFIS_TEMPFAIL;
    }
    elsif ($ok) {
	# add a header to the message announcing our presence
	warn "addhedr [$ctx] [interp#$ctx->{__perlmx_id}]\n";
	$ctx->addheader("X-Archived", $ctx->{fname} . '@' . hostname());
    }
    else {
	unlink($ctx->{fname});
    }
    return $status;
}

# callbacks

sub cb_connect {
    my $ctx = shift;
    my $name = shift;
    warn "connect [$ctx] [interp#$ctx->{__perlmx_id}] [$name]\n";
    return SMFIS_CONTINUE;
}

sub cb_envfrom {
    my $ctx = shift;
    warn "envfrom [$ctx] [interp#$ctx->{__perlmx_id}]\n";
    @{$ctx}{'fname','fh'} = make_temp_file() or return SMFIS_TEMPFAIL;
    return SMFIS_CONTINUE;
}

sub cb_header {
    my $ctx = shift;
    my $headerf = shift;
    my $headerv = shift;
    printf { $ctx->{fh} } "%s: %s\r\n", $headerf, $headerv;
    return SMFIS_CONTINUE;
}

sub cb_eoh {
    my $ctx = shift;
    print { $ctx->{fh} } "\r\n";
    return SMFIS_CONTINUE;
}

sub cb_body {
    my $ctx = shift;
    my $body = shift;
    print { $ctx->{fh} } $body;
    return SMFIS_CONTINUE;
}

sub cb_eom {
    my $ctx = shift;
    return $ctx->cleanup(1);
}

sub cb_abort {
    my $ctx = shift;
    return $ctx->cleanup(0);
}

sub cb_close {
    return SMFIS_ACCEPT;
}

sub new {
    my $class = shift;
    my $s = {
		NAME	=> "PerlMx::Samples::SampleFilter",
		FLAGS	=> SMFIF_ADDHDRS,
		CONNECT	=> \&cb_connect,
		ENVFROM	=> \&cb_envfrom,
		HEADER	=> \&cb_header,
		EOH	=> \&cb_eoh,
		BODY	=> \&cb_body,
		EOM	=> \&cb_eom,
		ABORT	=> \&cb_abort,
		CLOSE	=> \&cb_close,
	    };
    return bless $s, $class;
}

warn "Loading " . __PACKAGE__ . "\n";

1;

__END__

=head1 NAME

PerlMx::Samples::SampleFilter - Logs messages

=head1 SYNOPSIS

    pmx1 PerlMx::Samples::SampleFilter >& SampleFilter.log &

=head1 DESCRIPTION

This is a sample filter that logs messages in the /tmp directory.
Each message logged has an additional X-Archived header that specifies
the file name to which it has been logged.

=head1 HIGH-LEVEL OVERVIEW

This filter overrides the mail filter base class by defining a function
called new, which creates a hash containing data, and blesses it into the class
name. Among other things, the hash contains references to the callback functions
that the base class should call. 

=head1 DETAILS

This section gives as much detail as possible about the code.

=head2 Preamble

    package PerlMx::Samples::SampleFilter; 

This line lets Perl know that this file provides the package SampleFilter. 

    use 5.006
    use strict;
    use warnings;

This chunk gives Perl instructions about how to compile the code. For
example, use 5.006 means that this module must be compiled with Perl
5.006 or above. use strict and use warnings control how picky the
compiler is about warning for things like undeclared variables, and so
on.

    use PerlMx;
    use base qw(PerlMx);

This imports the base class into the filter.

    use Sys::Hostname;
    use Fcntl;

These are system packages that the filter requires. Sys::Hostname is
used to import the function hostname(), which returns the name of the
computer this code is running on. Fcntl imports sysopen(), used to open
files on the disk.

    sub make_temp_file {
        my $seq = int rand(10000000);
        my $file = "/tmp/smfi$seq";
        my $fh;
        my $iter = 0;
        my $flags = O_CREAT|O_EXCL|O_RDWR|O_BINARY;
        # XXX Real code should use File::Temp::mkstemp() but that's not
        # available in Perl v5.6.0.
        while (!sysopen($fh, $file, $flags, 0600)) {
            $seq = int rand(10000000);
            $file = "/tmp/smfi$seq";
            return if ++$iter > 10;
        }
        return ($file,$fh);
    }

This subroutine generates a random filename and attempts to open it
for writing. In case a file by that name exists, it tries up to 10
random filenames before failing. If successful, it returns the
filename and file handle.

    sub cleanup {
        my $ctx = shift;
        my $ok = shift;
        my $status = SMFIS_CONTINUE;
        if (exists $ctx->{fh} and !close($ctx->{fh})) {
            unlink($ctx->{fname});
            $status = SMFIS_TEMPFAIL;
        }
        elsif ($ok) {
            # add a header to the message announcing our presence
            warn "addhedr [$ctx] [interp#$ctx->{__perlmx_id}]\n";
            $ctx->addheader("X-Archived", $ctx->{fname} . '@' . hostname());
        }
        else {
            unlink($ctx->{fname});
        }
        return $status;
    }

cleanup() is passed a reference to the SampleFilter object when it is
called. It closes the file which was created by make_temp_file(), and if that
fails, it returns an error. If not, it checks whether it was called to 
add an extra header announcing that the message was logged. If it was not
called to with this parameter, it simply deletes the file.

=head2 Callbacks

    sub cb_connect {
        my $ctx = shift;
        my $name = shift;
        warn "connect [$ctx] [interp#$ctx->{__perlmx_id}] [$name]\n";
        return SMFIS_CONTINUE;
    }

This subroutine is called on every incoming SMTP connection. The first
argument (not counting $ctx) is the hostname, as determined by a
reverse lookup on the host. The second argument is the host
address. In this case, cb_connect() prints an informative message, but
does no real processing.

    sub cb_envfrom {
        my $ctx = shift;
        warn "envfrom [$ctx] [interp#$ctx->{__perlmx_id}]\n";
        @{$ctx}{'fname','fh'} = make_temp_file() or return SMFIS_TEMPFAIL;
        return SMFIS_CONTINUE;
    }

This subroutine is called each time sendmail receives an "envelope
from" command from the SMTP transaction. The first argument (not
counting $ctx) is the sender address; other arguments, if any, are
ESMTP arguments. In this case, cb_envfrom() prints an informative
message and creates a temporary file.

    sub cb_header {
        my $ctx = shift;
        my $headerf = shift;
        my $headerv = shift;
        printf { $ctx->{fh} } "%s: %s\r\n", $headerf, $headerv;
        return SMFIS_CONTINUE;
    }

This callback is invoked on each header of the message. The first
argument is the name of the field (i.e. "To") and the second is the
value of that field (i.e. "postmaster@localhost.localdomain"). In this
case, the header is logged to the temporary file.

    sub cb_eoh {
        my $ctx = shift;
        print { $ctx->{fh} } "\r\n";
        return SMFIS_CONTINUE;
    }

This subroutine is called after all the headers in the message. In
this case, it prints an extra newline to the temporary file, after the
end of all the headers.

    sub cb_body {
        my $ctx = shift;
        my $body = shift;
        print { $ctx->{fh} } $body;
        return SMFIS_CONTINUE;
    }

This subroutine is called for each chunk of the body message. This
routine may be called more than once for each message, as the body may
contain multiple chunks. In this case it prints the body to the
temporary file.

    sub cb_eom {
        my $ctx = shift;
        return $ctx->cleanup(1);
    }

This subroutine is called at the end of the message. If this message
is called, cb_abort will not be called. In this case, cb_eom() calls the
cleanup() function with arguments to close the temporary file, and add
a header to the message announcing that it was logged to a temporary
file.

    sub cb_abort {
        my $ctx = shift;
        return $ctx->cleanup(0);
    }

This callback is invoked if the message is aborted somewhere outside
the filter (for example, when a QUIT message is sent before the
message is complete). In this case, it calls cleanup with instructions
to delete the temporary file.

    sub cb_close {
        return SMFIS_ACCEPT;
    }

This callback is invoked at the end of the connection. 

=head2 Constructor

    sub new {
        my $class = shift;
        my $s = {
                    NAME	=> "PerlMx::Samples::SampleFilter",
                    FLAGS	=> SMFIF_ADDHDRS,
                    CONNECT	=> \&cb_connect,
                    ENVFROM	=> \&cb_envfrom,
                    HEADER	=> \&cb_header,
                    EOH	=> \&cb_eoh,
                    BODY	=> \&cb_body,
                    EOM	=> \&cb_eom,
                    ABORT	=> \&cb_abort,
                    CLOSE	=> \&cb_close,
                };
        return bless $s, $class;
    }

This function is the constructor of the class. It creates a hash containing 
the name of the filter, the flags, and which callback functions to invoke, 
and it blesses this structure into an object named $class. It returns this
blessed object.

=cut
