#
# A sample to illustrate a filter in action.
#
# This is the filter implementation.  See also pmx1 (the filter
# driver).
#

package PerlMx::Samples::SimpleRedirect;

use 5.006;
use strict;
use warnings;

use PerlMx;
use base qw(PerlMx);

# Lexical scope
my %name_changes;
my $database_file = "SimpleRedirect.conf";
open DBS, $database_file;
flock (DBS, 1); # shared lock
my @lines = <DBS>;
flock (DBS, 8); # unlock it!
close DBS;

my %lookup;
foreach (@lines) {
  my @first = split /:\s*/;
  next if (!$#first);
  chomp (my @names = split /,\s*/, $first[1]);
  $lookup{$first[0]} = \@names;
}

# utility functions

sub get_fw_rcpt {
  my $rcpt = shift;
  my $debug = shift;

  my %final;
  my $dork = get_names($debug,$rcpt,\%lookup, (), ());
  my @names = sort keys %$dork;
  if ($debug) { print "these terminals were found: [@names]\n"; }
  return @names;
}

sub get_names {
  my ($debug, $rcpt,$names,$terminals,$non_terminals) = @_;

  my $name;
  if (exists $names->{$rcpt}) {
    foreach $name (@{$names->{$rcpt}}) {
      if (exists $names->{$name}) {
	print "$name: non_terminal; " if $debug;

	if (exists $non_terminals->{$name}) {
	  print "previously parsed... stopping\n" if $debug;
	} else {
	  print "not previously parsed... recursing\n" if $debug;
	  $non_terminals->{$name} = 0;

	  # just because $terminals may be as-yet undefined, we have to
	  # assign it here to make sure we don't lose the results of the
	  # recursion.
	  $terminals = get_names($debug,
				 $name,
				 $names,
				 $terminals,
				 $non_terminals);
	}
      } else {
	print "$name: terminal name... stopping\n" if $debug;
	$terminals->{$name} = 0;
      }
    }
  }

  return $terminals;
}

sub cleanup {
    my $ctx = shift;
    my $ok = shift;
    my $status = SMFIS_CONTINUE;

    return $status;
}

# callbacks

sub cb_envrcpt {
    my $ctx = shift;
    my $rcpt = shift;

    $rcpt =~ /[^<]*<?([^>]+)>?/;
    $1 =~ /([^@]+)@/;
    my @fw_rcpt = get_fw_rcpt($1, 0);
    $name_changes{$rcpt} = \@fw_rcpt unless ($#fw_rcpt < 0);

    return SMFIS_CONTINUE;
}

sub cb_eom {
    my $ctx = shift;

    foreach (sort keys %name_changes) {
      print "deleting recipient: $_\n";
      $ctx->delrcpt($_);
      foreach (@{$name_changes{$_}}) {
	print "adding recipient: $_\n";
	$ctx->addrcpt($_);
      }
    }

    return $ctx->cleanup(1);
}

sub cb_abort {
    my $ctx = shift;
    return $ctx->cleanup(0);
}

sub cb_close {
    return SMFIS_ACCEPT;
}

sub new {
    my $class = shift;
    my $s = {
		FLAGS	=> SMFIF_ADDRCPT | SMFIF_DELRCPT,
	        ENVRCPT => \&cb_envrcpt,
		EOM	=> \&cb_eom,
		ABORT	=> \&cb_abort,
		CLOSE	=> \&cb_close,
	    };
    return bless $s, $class;
}

warn "Loading " . __PACKAGE__ . "\n";

1;

__END__

=head1 NAME

PerlMx::Samples::SimpleRedirect - Redirects mail

=head1 SYNOPSIS

    pmx1 PerlMx::Samples::SimpleRedirect >& SimpleRedirect.log &

=head1 DESCRIPTION

This is a sample filter that redirects mail based on a flat
database similar to (but more limited than) the aliases file used by
sendmail itself. The filename is defined in SimpleRedirect.pm, just
for an example.

=head1 HIGH-LEVEL OVERVIEW

It is possible that in some cases it would be desirable to
automatically redirect mail from one user to another user on the
fly. This simple filter demonstrates this ability in its
simplest form. In this case, names are stored in a text file of 
the following form:

  name: alternate1,alternate2,alternate3,...
  name2: ...
  ...

While this is extremely simple, it is functional. A simple advancement
would be to use a real database, so that the database could really
be updated on-the-fly. In addition, databases have more advanced locking
capabilities than text-files.

=head1 DETAILS

This section gives as much detail as possible about the code.

=head2 Preamble

  package PerlMx::Samples::SimpleRedirect;

This line tells Perl that this filter lives in the package
PerlMx::Samples::SimpleRedirect.

  use 5.006;
  use strict;
  use warnings;

This chunk gives Perl instructions about how to compile the code. For
example, use 5.006 means that this module must be compiled with Perl
5.006 or above. use strict and use warnings control how picky the
compiler is about warning for things like undeclared variables, and so
on.

  use PerlMx;
  use base qw(PerlMx);

This imports the base class into the filter.

  # Lexical scope
  my %name_changes;
  my $database_file = "SimpleRedirect.dbs";

These lines declare two lexically-scoped variables: %name_changes is a
hash, and $database_file is a string containing the name of the
redirection database file.

  open DBS, $database_file;
  flock (DBS, 1);
  my @lines = <DBS>;
  flock (DBS, 8);
  close DBS;

These lines import the database into the array @lines, which is also
lexically-scoped. While it reads the database, it locks the file.

  my %lookup;
  foreach (@lines) {
    my @first = split /:\s*/;
    next if (!$#first);
    chomp (my @names = split /,\s*/, $first[1]);
    $lookup{$first[0]} = \@names;
  }

This block of code parses the database into a graph structure, so that
each key in the hash %lookup contains an array of names. If mail
arrives for someone whose name is a key in the hash, it means that
their mail should be sent to those people in the array. 

  sub get_fw_rcpt {
    my $rcpt = shift;
    my $debug = shift;

    my %final;
    my $dork = get_names($debug,$rcpt,\%lookup, (), ());
    my @names = sort keys %$dork;
    if ($debug) { print "these terminals were found: [@names]\n"; }
    return @names;
  }

This function, get_fw_rcpt(), is called to determine whether a
particular user should be forwarded to anyone. If the user's mail should
be forwarded, then the function returns an array of alternate addresses.

  sub get_names {
    my ($debug, $rcpt,$names,$terminals,$non_terminals) = @_;

    my $name;
    if (exists $names->{$rcpt}) {
      foreach $name (@{$names->{$rcpt}}) {
        if (exists $names->{$name}) {
	  print "$name: non_terminal; " if $debug;

  	  if (exists $non_terminals->{$name}) {
	    print "previously parsed... stopping\n" if $debug;
	  } else {
	    print "not previously parsed... recursing\n" if $debug;
	    $non_terminals->{$name} = 0;

	    # just because $terminals may be as-yet undefined, we have to
	    # assign it here to make sure we don't lose the results of the
	    # recursion.
	    $terminals = get_names($debug,
				   $name,
				   $names,
				   $terminals,
				   $non_terminals);
	  }
        } else {
	  print "$name: terminal name... stopping\n" if $debug;
	  $terminals->{$name} = 0;
        }
      }
    }

    return $terminals; 
  }

get_names() recursively traverses the hash of names, trying to
eliminate situations where a substitute recipient is also supposed to
have their mail forwarded to someone else. For example, if Bob has his
mail forwarded to Jane, and Jane also has her mail forwarding to Mary,
then this function will decide to send Bob's email to Mary.

  sub cleanup {
      my $ctx = shift;
      my $ok = shift;
      my $status = SMFIS_CONTINUE;

      return $status;
  }

This function doesn't do anything in this case, but it could be
extended to clean up anything that needed cleaning up later. 

=head2 Callbacks

  sub cb_envrcpt {
      my $ctx = shift;
      my $rcpt = shift;

      $rcpt =~ /[^<]*<?([^>]+)>?/;
      $1 =~ /([^@]+)@/;
      my @fw_rcpt = get_fw_rcpt($1, 0);
      $name_changes{$rcpt} = \@fw_rcpt unless ($#fw_rcpt < 0);

      return SMFIS_CONTINUE;
  }

This callback method is invoked on every envelope recipient for the 
SMTP transaction. The first argument (except $ctx) is the recipient
address. In this case, the recipient's address is stripped to just
the bare-word username, and this user address is the one which is
compared to the names in the database via get_fw_rcpt(). If no names
are returned by get_fw_rcpt(), then nothing happens. If names were found,
they are saved in the %name_changes hash for later use.

  sub cb_eom {
      my $ctx = shift;

      foreach (sort keys %name_changes) {
        print "deleting recipient: $_\n";
        $ctx->delrcpt($_);
        foreach (@{$name_changes{$_}}) {
	  print "adding recipient: $_\n";
	  $ctx->addrcpt($_);
        }
      }

      return $ctx->cleanup(1);
  }

This callback is invoked at the end of the message. In this case, it
steps through each name in the %name_changes hash and deletes it from
the recipients of the message, and adds each alternate recipient to
the recipient list.

  sub cb_abort {
      my $ctx = shift;
      return $ctx->cleanup(0);
  }

This callback can be used to reset the state. In this case, it does
nothing.

  sub cb_close {
      return SMFIS_ACCEPT;
  }

This callback is called each time the SMTP connection is closed. It
does nothing here.

=cut

