#
# A sample to illustrate a filter in action.
#
# This is the filter implementation.  See also pmx1 (the filter
# driver).
#

package PerlMx::Samples::SimpleWordFilter;

use 5.006;
use strict;
use warnings;

use PerlMx;
use base qw(PerlMx);

use Fcntl ':flock';

use Sys::Hostname;
use Data::Dumper;
use File::Temp;

# variables
my $config_file="SimpleWordFilter.conf";	# name of the config file and word list
my %badwords;

init();

sub init {
  print "INIT()\n";

  open INIT, $config_file;
  flock (INIT, LOCK_SH);
  my @t = <INIT>;
  close INIT;

  my $act_ref;
  foreach (@t) {
    next if (/^\s*\#/ || /^\s*$/);

    if (/ACTION/) {
      chomp;
      my @args = split /:/g;
      $act_ref = [ @args[1..$#args] ];
      print "'@args' ... '@args[1..$#args]'\n";
    }

    else {
      chomp;
      $badwords{$_} = $act_ref;
    }
  }

  print Dumper(\%badwords);

}

# utility functions

sub cleanup {
    my $ctx = shift;
    my $ok = shift;
    my $status = SMFIS_CONTINUE;

    return $status;
}

# callbacks

sub cb_connect {
  my $ctx = shift;

  $ctx->{cbody} = "";

  return SMFIS_CONTINUE;
}

sub cb_envfrom {
    my $ctx = shift;

    $ctx->{sender} = shift;

    return SMFIS_CONTINUE;
}

sub cb_envrcpt {
    my $ctx = shift;
    my $rcpt = shift;

    # add the latest recipient
    push (@{$ctx->{rcpts}}, $rcpt);

    return SMFIS_CONTINUE;
}

sub cb_header {
    my $ctx = shift;
    my $headerf = shift;
    my $headerv = shift;

    push (@{$ctx->{headers}}, [ $headerf, $headerv ]);

    return SMFIS_CONTINUE;
}

sub cb_body {
    my $ctx = shift;
    my $body = shift;

    $ctx->{cbody} = $ctx->{cbody} . $body;

    if (exists $ctx->{action}) {}
    else {
      foreach (keys %badwords) {
	if ($body =~ /$_/i) {
	  $ctx->{action} = $badwords{$_};
	  last;
	}
      }
    }

    return SMFIS_CONTINUE;
}

sub cb_eom {
    my $ctx = shift;
    my $result = SMFIS_ACCEPT;

    if (exists $ctx->{action}) {
      my $action = delete $ctx->{action};
      if ($action->[0] eq "discard") {
	print "discarding\n";
	$result = SMFIS_DISCARD;
      }
      elsif ($action->[0] eq "forward") {
	print "forwarding to $action->[1]\n";
	foreach (@{$ctx->{rcpts}}) {
	  $ctx->delrcpt($_);
	}
	$ctx->addrcpt($action->[1]);
      }
      elsif ($action->[0] eq "logtofile") {
	print "logging to file\n";
	open LOG, "\>>$action->[1]" or
	  warn "couldn't open file $action->[1] for writing.\n";
	flock (LOG, LOCK_EX);
	foreach (@{$ctx->{headers}}) { print LOG "$$_[0]: $$_[1]\r\n"; }
	print LOG "\r\n";
	print LOG $ctx->{cbody};
	close LOG;
      }
      elsif ($action->[0] eq "logtodir") {
	my ($tmpfh,$tmpname) = File::Temp::mkstemp("$action->[1]/pmxXXXXXXXX");
	print "logging message to $tmpname\n";
	flock ($tmpfh, LOCK_EX);
	foreach (@{$ctx->{headers}}) { print $tmpfh "$$_[0]: $$_[1]\r\n"; }
	print $tmpfh "\r\n";
	print $tmpfh $ctx->{cbody};
	close $tmpfh;
      }
    }

    return $result;
}

sub cb_abort {
    my $ctx = shift;
    return $ctx->cleanup(0);
}

sub cb_close {
    return SMFIS_ACCEPT;
}

sub new {
    my $class = shift;
    my $s = {
	        CONNECT => \&cb_connect,
		ENVFROM	=> \&cb_envfrom,
		ENVRCPT	=> \&cb_envrcpt,
		HEADER	=> \&cb_header,
		BODY	=> \&cb_body,
		EOM	=> \&cb_eom,
		ABORT	=> \&cb_abort,
		CLOSE	=> \&cb_close,
	    };
    return bless $s, $class;
}

warn "Loading " . __PACKAGE__ . "\n";

1;

__END__

=head1 NAME

PerlMx::Samples::SimpleWordFilter - Scans message body for
keywords

=head1 SYNOPSIS

    pmx1 PerlMx::Samples::SimpleWordFilter >& SimpleWordFilter.log &

=head1 DESCRIPTION

A simple filter that scans the body for key words which are
read from a configuration file on startup. The configuration file
contains instructions for what should be done if a key word is found.

=head1 DEPENDENCIES

* File::Temp

=head1 COPYRIGHT

Copyright (c) 2000-2003 ActiveState. All rights reserved.

=head1 SEE ALSO

L<PerlMx>, L<PerlMx::Simple>, L<perl>

=cut
