#
# A sample to illustrate a filter in action.
#
# This is the filter implementation.  See also pmx1 (the filter
# driver).
#

package PerlMx::Samples::VolumeAccounting;

use 5.006;
use strict;
use warnings;

use PerlMx;
use base qw(PerlMx);

use Fcntl;
use Data::Dumper;

sub LOCK_SH() { 1 }
sub LOCK_EX() { 2 }

my $DEBUG = 0;

# utility functions
my $conffile = "VolumeAccounting.conf";
my $logfile = "volume-accounting.log";         # NOTE: must have permissions
my %options;

open (CONF, $conffile);
flock (CONF, LOCK_SH);
my @lines = <CONF>;
close (CONF);

foreach (@lines) {
    next if (/^\s*\#/ || /^\s*$/);
    chomp;
    $options{$_} = 1;
}

sub add_log_entry {
    # adds a line: <date>:<from>:<hdrsize>:<bdysize>:<totsize>
    my ($ctx) = @_;
    my $date = scalar time;
    my $line;

    my ($from,$bdysize,$rcpts) = @{$ctx}{'from','bdysize','rcpts'};

    if ($options{Totals}) {
	warn "Totals\n" if $DEBUG;

	my $check = 1;
	my @current;
	open LOG, $logfile or $check = 0;
	if ($check) {
	    flock (LOG, LOCK_SH);
	    @current = <LOG>;
	    close LOG;
	}

	# add to the rest
	my %current;
	foreach (@current) {
	    my @t = split /:/;
	    $current{$t[0]} = $t[1];
	}

	foreach (@$rcpts) {
	    $current{$_} += $bdysize;
	}

	if (open LOG, ">$logfile") {
	    flock (LOG, LOCK_EX)
		or warn "Could not lock log file $logfile.\n";
	    foreach (sort keys %current) {
		print LOG "$_:$current{$_}\n";
	    }
	    close LOG;
	}
	else {
	    warn "Could not open log file $logfile for writing: $!";
	}
    }
    elsif ($options{DailyTotals}) {
	warn "Daily Totals\n" if $DEBUG;

	my $check = 1;
	my @current;
	open LOG, $logfile or $check = 0;
	if ($check) {
	    flock (LOG, LOCK_SH);
	    @current = <LOG>;
	    close LOG;
	}

	my %b;
	my ($year,$day) = (localtime)[5,7];
	my $key = "$year.$day";

	foreach (@current) {
	    my @t = split /:/;
	    my $nkey = "$t[0]";
	    $b{"$nkey"}{$t[1]} = $t[2];
	}

	foreach (@$rcpts) {
	    $b{"$key"}{$_} += $bdysize;
	}

	warn Dumper(\%b) if $DEBUG;

	open LOG, ">$logfile" or warn "!\n";
	flock (LOG, LOCK_EX);
	foreach (keys %b) {
	    my $t;
	    foreach $t (keys %{$b{$_}}) {
		print LOG "$_:$t:$b{$_}{$t}\n";
	    }
	}
	close LOG;

    }
    elsif ($options{LogByRecipients}) {
	warn "Log by recips\n" if $DEBUG;

	open LOG, ">>$logfile" or warn "!\n";
	flock (LOG, LOCK_EX);
	foreach (@$rcpts) {
	    print LOG "$date:$_:$from:",$bdysize, "\n";
	}
	close LOG;
    }
    elsif ($options{LogByDay}) {
	warn "Log by day\n" if $DEBUG;
	# XXX not yet implemented
    }
    else {
	warn "No Logging actions!\n";
    }
}

sub cleanup {
    my $ctx = shift;
    my $ok = shift;
    my $status = SMFIS_CONTINUE;

    if (exists $ctx->{fh} and !close($ctx->{fh})) {
	close($ctx->{fname});
	$status = SMFIS_TEMPFAIL;
    }

    return $status;
}

# callbacks

sub cb_envfrom {
    my $ctx = shift;

    $ctx->{from} = shift;
    print ".\n";

    return SMFIS_CONTINUE;
}

sub cb_envrcpt {
    my $ctx = shift;

    push (@{$ctx->{rcpts}}, shift);
    print ".\n";

    return SMFIS_CONTINUE;
}

sub cb_body {
    my $ctx = shift;
    my $body = shift;

    $ctx->{bdysize} += length $body;

    print "$body";

    return SMFIS_CONTINUE;
}

sub cb_eom {
    my $ctx = shift;

    warn ".\n" if $DEBUG;
    add_log_entry($ctx);
    return $ctx->cleanup(1);
}

sub cb_abort {
    my $ctx = shift;
    return $ctx->cleanup(0);
}

sub cb_close {
    return SMFIS_ACCEPT;
}

sub new {
    my $class = shift;
    my $s = {
#		FLAGS	=> 0,
		ENVFROM	=> \&cb_envfrom,
	        ENVRCPT => \&cb_envrcpt,
		BODY	=> \&cb_body,
		EOM	=> \&cb_eom,
		ABORT	=> \&cb_abort,
		CLOSE	=> \&cb_close,
	    };
    return bless $s, $class;
}

warn "Loading " . __PACKAGE__ . "\n" if $DEBUG;

1;

__END__

=head1 NAME

PerlMx::Samples::VolumeAccounting - Logs messages

=head1 SYNOPSIS

    pmx1 PerlMx::Samples::VolumeAccounting >& VolumeAccounting.log &

=head1 DESCRIPTION

This is a sample filter that logs the size of each message
in a log file. A simple filter that counts the number of
bytes in each message (including headers) and logs a summary in a
log file.

=head1 COPYRIGHT

Copyright (c) 2000-2003 ActiveState. All rights reserved.

=head1 SEE ALSO

L<PerlMx>, L<perl>

=cut
