##############################################################################
#
# PerlMx::Samples::Word2HTML
#
# 
# A ready to use PerlMx filter that converts Microsoft Word documents to HTML
# and allows you to determine the fate of the original document.
#
# Michael Smith <mikesm@ActiveState.com>
#
# Copyright (c) 2000-2003 ActiveState. All rights reserved.
#
##############################################################################
package PerlMx::Samples::Word2HTML;

use strict;
use PerlMx::MIME_Filter;
use Sys::Hostname;
use base qw ( PerlMx::MIME_Filter );

$SIG{PIPE} = 'IGNORE';

our $DEBUG = 1;
our $config_file = shift || 'PerlMx-Word2HTML.conf';
my  %config = parse_config();

#
# Notification messages for each possible action
#
my  $quarantine_msg = <<EOT;
The original document has been quarantined and replaced by this HTML version.
The original document can be retrieved from:
    $config{'quarantine url'}quarantined.doc
EOT

my  $delete_msg = <<EOT;
The original document has been deleted and replaced by this HTML version.
EOT

my  $send_msg = <<EOT;
The original document has been converted to HTML and attached to the message.
EOT

#
# Constructor
#
sub new {
    my $class = shift;
    my $self = $class->SUPER::new(@_);
    $self->register_mimetype('raw', \&cb_raw);
    $self->register_mimetype('APPLICATION/msword', \&cb_application_msword);
    $self->output_under($config{'quarantine directory'});
    $self->purge(0) if $config{'action'} eq 'quarantine';
    $self->message_changed(0);
    return $self;
}

#
# Use the `raw' callback to store the top level MIME::Entity for later use
#
sub cb_raw {
    my $ctx = shift;
    my $e = shift;
   
    # Store a reference to the top level MIME::Entity for later use
    $ctx->{'top_entity'} = $e;
    return CONTINUE;
}

#
# Our workhorse
#
sub cb_application_msword {
    my $ctx = shift;
    my $e = shift;
    
    if ($config{'action'} eq 'quarantine') {
	quarantine($e);
	remove_part($ctx->{'top_entity'}, $e);
	notify($ctx->{'top_entity'}, $quarantine_msg);
	attach_HTML($ctx->{'top_entity'}, $e);
    }
    elsif ($config{'action'} eq 'delete') {
	remove_part($ctx->{'top_entity'}, $e);
	notify($ctx->{'top_entity'}, $delete_msg);
	attach_HTML($ctx->{'top_entity'}, $e);
    }
    elsif ($config{'action'} eq 'send') {
	notify($ctx->{'top_entity'}, $send_msg);
	attach_HTML($ctx->{'top_entity'}, $e);
    }
    else {
	warn "Unknown action $config{'action'}";
    }
    $ctx->message_changed(1);

    return CONTINUE;
}

#
# Remove the MIME::Entity referenced by $e. $e must be decendent from $top
#
sub remove_part {
    my $top = shift;
    my $e   = shift;
    
    $top->parts( [ grep {!/\Q$e\E/} $top->parts ] );
}

#
# Simple quarantine function. Hopefully in the future this will call a 
# standard method of some quarantine object or a function in PerlMx::Utils
#
sub quarantine {
    my $e = shift;
    $e->print(\*STDERR);
}

#
# Add a simple notification message to the provided MIME::Entity object.
# In the future this should call a function in PerlMx::Utils
#
sub notify {
    my $e = shift;
    my $msg = shift;
    $e->attach(Data => $msg,
	       Type => 'text/plain');
}

#
# Convert the attached Word document to HTML and attach the HTML to the 
# top level MIME::Entity
#
sub attach_HTML {
    my $top = shift;
    my $e = shift;
    my $file;
    my @html;

    if ($e->bodyhandle && ($file = $e->bodyhandle->path)) {
	open HTML, "$config{'wv program'} --config $config{'wv config'} $file |" or warn "$config{'wv program'} $file: $!";
	@html = <HTML>;
	$top->attach(Data	    => join('', @html),
		     Type	    => 'text/html',
		     Disposition    => 'attachment');
	close HTML if @html;
    }
    else {
	warn 'Body is not stored on disk';
    }
}    

#
# Get our configuration parameters
#
sub parse_config {
    my %config;
    open CONF, $config_file or warn "$config_file : $!";
    while (<CONF>) {
	if (/^#|^\s*$/) { 
	    next; 
	}
	elsif (/^wv program\s+(.*)$/) {
	    $config{'wv program'} = $1;
	}
	elsif (/^wv config\s+(.*)$/) {
	    $config{'wv config'} = $1;
	}
	elsif (/^action\s+(.*)$/) {
	    $config{'action'} = $1;
	}
	elsif (/^quarantine url\s+(.*)$/) {
	    $config{'quarantine url'} = $1;
	}
	elsif (/^quarantine directory\s+(.*)$/) {
	    $config{'quarantine directory'} = $1;
	}
	else {
	    warn 'Unknown configuration option ' . $_;
	}
    }

    close CONF if keys %config;

    # Default configuration parameters
    $config{'wv program'}	    ||= 'wvHtml';
    $config{'wv config'}	    ||= 'wvHtml.xml';
    $config{'action'}		    ||= 'send';
    $config{'quarantine url'}	    ||= 'http://' . hostname() . '/';
    $config{'quarantine directory'} ||= '/tmp';

    # return configuration hash
    return %config;
}

1;
__END__

=head1 NAME

PerlMx::Samples::Word2HTML - Converts
MS Word documents to HTML

=head1 SYNOPSIS

    pmx1 PerlMx::Samples::Word2HTML >& Word2HTML.log &

=head1 DESCRIPTION

=head2 CONFIGURATION

=over 4

=item wv program:

The location of the `wv' program. If `wv' is in the path then just the file
name is required. Otherwise, the absolute path to `wv' is required. wv is
actually a suite of programs for manipulating and converting MS Word documents.
The specific program that converts MS Word documents to HTML is called
`wvHtml'.

The `wv' program can be downloaded from:

      http://www.wvware.com/

wv program /usr/home/mikesm/working/wv/wvHtml
      
The project source is located at:

      http://sourceforge.net/projects/wvware/

=item wv config:

wv config /usr/home/mikesm/working/wv/wvHtml.xml

=item action:

One of: 

	quarantine, delete, send 

action send

=item quarantine url:

Base url from which quarantined documents can be retrieved

quarantine url http://clamp.ActiveState.com/

=item quarantine directory:

The absolute path to the directory in which documents are to be quarantined.

quarantine directory /tmp/quarantine

=back

=head1 COPYRIGHT

Copyright (c) 2000-2003 ActiveState. All rights reserved.

=head1 SEE ALSO

L<perl>, L<PerlMx>, http://www.wvware.org
