###############################################################################
#
# package PerlMx::MIME_Filter;
#
# Base class for MIME filters.
#
# Michael Smith <mikesm@ActiveState.com>
#
###############################################################################

package PerlMx::MIME_Filter;

use strict;
use MIME::Parser;
use IO::File;
use File::Path;
use PerlMx;
use base qw ( Exporter PerlMx );

our $VERSION = '3.000';

###############################################################################
#
# Exports
#
###############################################################################
our @EXPORT = qw(
        ACCEPT
        CONTINUE
        DISCARD
        REJECT
        TEMPFAIL
);

sub ACCEPT ()   { SMFIS_ACCEPT }
sub CONTINUE () { SMFIS_CONTINUE }
sub DISCARD ()  { SMFIS_DISCARD }
sub REJECT ()   { SMFIS_REJECT }
sub TEMPFAIL () { SMFIS_TEMPFAIL }

###############################################################################
#
# Globals
#
###############################################################################
our $DEBUG = 0;
our $SEQ = 0;
my  $tmp = get_temp();

###############################################################################
#
# Public Interface
#
###############################################################################

#
# Constructor
#
sub new {
    
    my $proto = shift;
    my $class = ref $proto || $proto;
    my $self  = ref $proto ? $proto : {};

    warn 'Creating new ' . $class . "\n" if $DEBUG > 3;

    $self = bless {
		CONNECT	=> 'pmx_connect',
		HELO	=> 'pmx_helo',
		ENVFROM => 'pmx_envfrom',
		ENVRCPT	=> 'pmx_envrcpt',
		HEADER  => 'pmx_header',
		EOH     => 'pmx_eoh',
		BODY    => 'pmx_body',
		EOM     => 'pmx_eom',
		ABORT   => 'pmx_abort',
		CLOSE	=> 'pmx_close',
	    }, $class;

    $self->purge(1);
    $self->message_changed(1);

    return $self;
}

sub output_under {
    my $ctx = shift;
    my $dir = shift;
    if (-d $dir) {
	$ctx->{__perlmxm_temp_dir} = $dir;
    }
    else {
	warn "$dir does not exist\n";
    }
}

sub purge {
    my $ctx = shift;
    my $old = $ctx->{__perlmxm_purge};
    $ctx->{__perlmxm_purge} = shift if @_;
    return $old;
}

sub message_changed {
    my $ctx = shift;
    my $old = $ctx->{__perlmxm_msg_changed};
    $ctx->{__perlmxm_msg_changed} = shift if @_;
    return $old;
}

sub register_mimetype {
    my $ctx = shift;
    my $type = shift;
    my $c = shift;

    if (ref $c ne 'CODE') {
	warn "Not a code reference in register_mimetype()";
	return 0;
    }

    warn "Registered callback $c for MIME-Type $type\n" if $DEBUG > 3;
    push @{$ctx->{__perlmxm_handlers}->{$type}}, $c;
    return 1;
}

sub filter {
    warn "Filtering\n" if $DEBUG > 3;
    my $ctx = shift;
    my $p = MIME::Parser::->new();
    my $entity;
    my $status = CONTINUE;
    my $called_cb;  # have we called any callback

    my @mime_types = keys %{$ctx->{__perlmxm_handlers}};
   
    seek( $ctx->{fh}, 0, 0 );
    my $tmpdir = $ctx->{__perlmxm_temp_dir} || get_temp();
    my $subdir = "msg-" . scalar(time) . "-$$-$ctx->{__perlmx_id}-" . ++$SEQ;
    $p->output_under($tmpdir, "DirName" => $subdir);
    $ctx->{__perlmxm_me} = $p->parse( $ctx->{fh} );
    
    if($ctx->{__perlmxm_me}) {
	$ctx->{__perlmxm_output_dir} = $p->output_dir;
	
	# Always run `raw' if it is defined
	if (defined $ctx->{__perlmxm_handlers}->{raw}) {
	    warn "Calling `raw' handler\n" if $DEBUG > 3;
	    foreach my $cb (@{$ctx->{__perlmxm_handlers}->{raw}}) {
		$status = $cb->( $ctx, $ctx->{__perlmxm_me} );
		$called_cb++;
		last if $status != CONTINUE;
	    }
	}

	# See if we have a handler for each of the MIME entities
	# contained in this message
      PART:
	foreach my $e ($ctx->{__perlmxm_me}->parts_DFS) {

	    warn 'Found ' . $e->effective_type . " part\n" if $DEBUG > 3;
	    
	    foreach my $mime_type (@mime_types) {
		warn 'Calling `' . $e->mime_type . "\' handler\n" if $DEBUG > 3;

		if($e->effective_type =~ /^$mime_type$/i) {
		    foreach my $cb (@{$ctx->{__perlmxm_handlers}->{$mime_type}}) {
			$status = $cb->( $ctx, $e );
			$called_cb++;
			last PART if $status != CONTINUE;
		    }
		}
	    }
	}

	if ($called_cb && $ctx->message_changed &&
	    ($status == CONTINUE || $status == ACCEPT))
	{
	    # Write back the modified message to sendmail
	    my $mime_entity = $ctx->{__perlmxm_me};
	    my $content_type = $mime_entity->head->get('Content-Type');
	    my $mime_version = $mime_entity->head->get('MIME-Version');
	    for ($content_type, $mime_version) {
		$_ = "" unless defined $_;
		$_ =~ s/\r?\n\z//;
	    }
	    
	    $status = ACCEPT;

	    tie ((local *H), 'PerlMx::MIME_Filter::Handle', $ctx);
	    $ctx->chgheader('MIME-Version', 1, $mime_version);
	    $ctx->chgheader('Content-type', 1, $content_type);
	    $mime_entity->print_body(\*H);
	}

	$ctx->clean_up;
    }
    else {
	warn "Could not parse message", $p->last_error, $p->results->errors, $p->results->top_head;
    }
    
    return $status;
}

# utility functions

sub get_temp {
    my ($tmp) = $ENV{TMP} || $ENV{TEMP} || '/tmp';
    warn "$tmp directory does not exist" if ! -d $tmp;
    return $tmp;
}

sub clean_up {
    my $ctx = shift;
    if($ctx->{__perlmxm_me}) {
    	$ctx->{__perlmxm_me}->purge if $ctx->purge;
    	delete $ctx->{__perlmxm_me};
    }

    (rmtree($ctx->{__perlmxm_output_dir}) 
	or warn "Could not rmtree($ctx->{__perlmxm_output_dir})!")
	    if $ctx->purge;
}

###############################################################################
# 
# Callbacks
#
###############################################################################
sub pmx_connect {
    return CONTINUE;
}

sub pmx_helo {
    return CONTINUE;
}

sub pmx_envfrom {
    my $ctx = shift;
    $ctx->{__pmxm_from} = shift;
    warn "envfrom [$ctx] [interp#$ctx->{__perlmx_id}]\n" if $DEBUG > 3;
    $ctx->{fh} = IO::File->new_tmpfile() or return TEMPFAIL;
    return CONTINUE;
}

sub pmx_envrcpt {
    return CONTINUE;
}

sub pmx_header {
    my $ctx = shift;
    my $headerf = shift;
    my $headerv = shift;
    printf { $ctx->{fh} } "%s: %s\r\n", $headerf, $headerv;
    return CONTINUE;
}

sub pmx_eoh {
    my $ctx = shift;
    print { $ctx->{fh} } "\r\n";
    return CONTINUE;
}

sub pmx_body {
    my $ctx = shift;
    my $body = shift;
    print { $ctx->{fh} } $body;
    return CONTINUE;
}

sub pmx_eom {
    my $ctx = shift;
    my $rv = CONTINUE;
    eval {
	$rv = $ctx->filter();
    };
    warn $@ if $@;
    if ($ctx->{fh}) {
	close $ctx->{fh};
	delete $ctx->{fh};
    }
    return $rv;
}

sub pmx_abort {
    my $ctx = shift;
    if ($ctx->{fh}) {
	close $ctx->{fh};
	delete $ctx->{fh};
    }
    return TEMPFAIL;
}

sub pmx_close {
    return CONTINUE;
}

###############################################################################
# 
# Tie::Handle object used in case MIME::Entity can take advantage of it
#
###############################################################################
package PerlMx::MIME_Filter::Handle;

require Tie::Handle;
our @ISA = qw/ Tie::Handle /;

sub TIEHANDLE {
    my $class = shift;
    my $self = { ctx => shift };
    bless $self, $class;
}

sub WRITE {
    my $self = shift;
    my $buf = shift;
    my $len = shift;
    my $off = shift;
    $self->{ctx}->replacebody(substr($buf, $off, $len));
}

sub PRINT {
    my $self = shift;
    $self->{ctx}->replacebody(join('', @_));
}

sub PRINTF {
    my $self = shift;
    my $fmt  = shift;
    $self->{ctx}->replacebody(sprintf($fmt, @_));
}

1;

__END__

=head1 NAME

PerlMx::MIME_Filter - Filter for MIME messages

=head1 SYNOPSIS

    package MyFilter;
    use PerlMx::MIME_Filter;
    use base qw(PerlMx::MIME_Filter);

    sub new {
	my $class = shift;
	my $mf = $class->SUPER::new;

	$mf->register_mimetype('raw', \&pmx_raw);
	$mf->register_mimetype('text/plain', \&cb_text_plain);
	$mf->register_mimetype('text/html', \&cb_text_html);

	return $mf;
    }

    sub pmx_raw {
	my $ctx = shift;
	my $e = shift;

	warn "MIME-Type: " . $e->mime_type . 
		" Number of parts: " . $e->parts . "\n";
    }

    sub cb_text_plain {
	my $ctx = shift;
	my $e = shift;

	$e->signature('This message does not necessarily reflect 
			the views of this organization');
    }


    ...other callback definitions...

    1;		# standard end of module


=head1 DESCRIPTION

This module provides a base class for users who wish to write their own 
filters to filter MIME messages. It provides a simple to use interface for
registering handlers to be called when MIME entities of a specified MIME type
are encountered in a message. In addition you can set callbacks for all the 
message delivery phases supported by C<PerlMx>.

The C<MIME-tools> distribution from CPAN must be installed before you
can use this module.

=head2 PerlMx::MIME_Filter Methods

=over 5

=item PerlMx::MIME_Filter->new( )

Your filter should provide a I<new()> class method that will call the base
class constructor, register one or more handlers and return a reference to
the newly created filter. Here is a sample:

    sub new {
	my $class = shift;
	my $mf = $class->SUPER::new; # Call the base class constructor

	# Register one or more MIME type handlers
	$mf->register_mimetype(...); 
	$mf->message_changed(0);

	return $mf;
    }

B<NOTE:> You must call the base class constructor. If you don't, some required
base class callbacks will not be registered, and filtering will not happen
correctly.

=item $mf->register_mimetype ( $mimetype, $code_ref )

Registers a handler which will be called when an entity of the specified 
MIME type is encountered. $mimetype is the MIME type for this handler. 
MIME types are matched case insensitively. $code_ref is a code reference to
the handler for the specified MIME type.

=item $mf->purge( $bool )

Allows the user to control whether the filter base class purges
temporary files from disk. If $bool is true (1), temporary files are
purged from disk at the end of the filter base class's EOM
callback. If $bool is FALSE (0), temporary files will not be purged
from disk. The default TRUE; i.e. to purge temporary files from disk

=item $mf->message_changed( $bool )

If this flag attribute is TRUE at the end of processing and the last
handler invoked has returned an ACCEPT or CONTINUE status, then the
message as seen by sendmail is also updated.  Handlers which modify
the MIME message and want the change to propegate to the mail
delivered should turn this flag on:

    $mf->message_changed(1);

For backwards compatibility reasons this flag starts out as TRUE.
The derived constructor should turn it off.

=item $mf->output_under( $dir )

Allows the user to control the directory under which the filter base class
stores temporary files while processing a message. $dir is the absolute path to
the directory under which temp files will be stored. The default directory is
determined in this order:

    $ENV{TMP} 
    $ENV{TEMP}
    /tmp

=back

B<NOTE:> All of the above methods, if called, should be called from the class
constructor, or some method called from the constructor.

=head2 PerlMx::MIME_Filter Handlers

Handlers are called if an entity of a registered MIME type is encountered in
the message. Handlers receive two parameters, a reference to the filter, see
L<PerlMx>, and a reference to the C<MIME::Entity> object that was encountered.
Using the C<MIME::Entity> object you can get access to and manipulate the 
corresponding C<MIME::Body>. Any changes made to the body will be reflected in
the message that is delivered to the recipients.

Handler return values should be one of the filter status codes documented in
L<PerlMx> or one of the alternative shorter names exported by this module.

A special handler can be set using the `raw' pseudo MIME type. If this handler
is set, it is guaranteed to be called once for each message and will be called
before any other handlers are called. This handler receives the same parameters
as a normal handler, the second parameter will be a reference to the top level
C<MIME::Entity> from which you can access all other C<MIME::Entity> objects
contained in the message.

B<Note:> if you need to manipulate the message headers, you should use the
filter object instead of the C<MIME::Entity> object. Any changes made to the
message headers in the C<MIME::Entity> object will be discarded.

=head2 EXPORT

The following constants are exported by default:

    ACCEPT
    CONTINUE
    DISCARD
    REJECT
    TEMPFAIL

=head1 COPYRIGHT

Copyright (c) 2000-2003 ActiveState Corp.  All rights reserved.

=head1 SEE ALSO

L<perl>, L<PerlMx>, L<MIME::Parser>, L<MIME::Entity>, L<MIME::Body>

=cut
