package PerlMx::Simple;
use PerlMx;
use base qw(Exporter PerlMx);
use strict;

our $VERSION = '3.000';

our @EXPORT = qw(
	ACCEPT
	CONTINUE
	DISCARD
	REJECT
	TEMPFAIL
);

sub ACCEPT ()	{ SMFIS_ACCEPT; }
sub CONTINUE ()	{ SMFIS_CONTINUE; }
sub DISCARD ()	{ SMFIS_DISCARD; }
sub REJECT ()	{ SMFIS_REJECT; }
sub TEMPFAIL ()	{ SMFIS_TEMPFAIL; }

sub new {
    my $pkg = shift;
    my $self = {};

    for my $cb (qw(connect helo envfrom envrcpt abort close)) {
	if (my $pmx_cb = $pkg->can("pmx_$cb")) {
	    $self->{uc $cb} = $pmx_cb;
	}
    }
    if (my $pmx_headers = $pkg->can("pmx_headers")) {
	$self->{EOH} = sub {
	    my $ctx = shift;
	    my $hdrs = shift;
	    my @tmp;
	    push(@tmp, join(": ", splice(@$hdrs, 0, 2))) while @$hdrs;
	    push(@tmp, "");
	    &$pmx_headers($ctx, join("\r\n", @tmp));
	};
    }
    if (my $pmx_body = $pkg->can("pmx_body")) {
	my $old_envfrom = $self->{ENVFROM};
	$self->{ENVFROM} = sub {
	    my $ctx = shift;
	    $ctx->{__perlmxs_body} = "";
	    return &$old_envfrom($ctx, @_) if $old_envfrom;
	    return SMFIS_CONTINUE;
	};
	$self->{BODY} = sub {
	    my $ctx = shift;
	    $ctx->{__perlmxs_body} .= $_[0];
	    return SMFIS_CONTINUE;
	};
	$self->{EOM} = sub {
	    my $ctx = shift;
	    &$pmx_body($ctx, $ctx->{__perlmxs_body});
	};
    }

    bless $self, $pkg;
}

1;
__END__

=head1 NAME

PerlMx::Simple - Simpler PureMessage interface

=head1 SYNOPSIS

    package MyFilter;     # customize name as needed
    use PerlMx::Simple;   # import constants
    use base qw(PerlMx::Simple);

    sub pmx_envfrom {
	my $ctx = shift;
	my($sender, @args) = @_;
	...process $sender and @args...
	return CONTINUE;  # or TEMPFAIL, REJECT etc.
    }

    ...other callback definitions...

    1;		          # standard end of module


=head1 DESCRIPTION

This module provides a simplified interface for writing C<PureMessage>
filters.  See L<PerlMx> for the more generic interface and additional
callback details.

C<PerlMx::Simple> provides much of the same functionality as C<PerlMx>,
except that it trades off some efficiency for a simpler interface.
There are three major simplifications:

=over

=item *

No constructor method needs to be defined.  If methods that have
well-known names are defined, they will be treated as filter
callbacks automatically.

=item *

The headers and body callbacks are passed all the headers and the
entire body in one go, respectively.  This means there are no
callbacks corresponding to the HEADER, EOH, BODY and EOM hooks
described in L<PerlMx>.

=item *

The status value constants do not have the C<SMFIS_> prefix.

=back

=head2 PerlMx::Simple Filter Methods

A C<PerlMx::Simple> filter module should start out as shown in the
SYNOPSIS section above.  In addition, the module should define
any additional callback methods as needed.  These methods have the
following "well-known" names and signatures.

=over

=item pmx_connect( $ctx, $hostname, $hostaddress )

This callback routine is invoked on every incoming SMTP connection.

=item pmx_helo( $ctx, $helo_host )

This callback routine is invoked on the SMTP HELO/EHLO command.

=item pmx_envfrom( $ctx, $sender, @esmtp_args )

This callback routine is invoked at the beginning of each message.

=item pmx_envrcpt( $ctx, $recipient, @esmtp_args )

This callback routine is invoked for every envelope recipient.

=item pmx_headers( $ctx, $headers )

This callback routine is invoked with the complete headers of the
message as a single multi-line string. Each header line is terminated
by Carriage-Return/Line-Feed.

=item pmx_body( $ctx, $msg_body )

This callback routine is invoked at the end of the message.  The
entire body of the message is passed as the $msg_body argument.
End-of-lines are represented by Carriage-Return/Line-Feed.

If this callback is called, C<pmx_abort> will not be called and vice versa.

This routine can perform special operations such as modifying the message
header, body, or envelope by calling one or more of the methods described
in L<PerlMx/"PerlMx Callback Methods">.

=item pmx_abort( $ctx )

This callback routine is invoked if the message is aborted outside of
the control of the filter.  This callback can be used to reset state.
If this callback is called, C<pmx_body> will not be called and vice
versa.

=item pmx_close( $ctx )

This callback method is invoked at the end of the SMTP connection.

=back

Callbacks can maintain state in hash object that is always passed as
their first argument ($ctx in the signatures above). The key namespace
that begins with C<__perlmx> is reserved for the base class to store
its own data.

Details about the arguments and behavior of C<pmx_connect>,
C<pmx_helo>, C<pmx_envfrom>, C<pmx_envrcpt>, C<pmx_abort> and
C<pmx_close> are found in the description of the corresponding method
in L<PerlMx/"Filter Callback Methods">.

=head2 Filter Status Codes

Each of the callback routines defined above may control how the SMTP
transaction is processed further by returning a status code.  Callbacks
should return one of following five status constants:

=over 10

=item CONTINUE

Continue processing the message or connection.  This is the normal return
value.

=item REJECT

Reject the message or connection.  No further filter callback routines will
be called.

=item DISCARD

Accept the message, but silently discard it.  No further routines will be
called for this message.

=item ACCEPT

Accept the message or connection.  No further routines will be called for
this message or connection.

=item TEMPFAIL

Return a temporary failure.

=back

Details about the effect of these status codes can be found in the
description of the corresponding codes in L<PerlMx/"Filter Status
Codes">.

=head1 EXAMPLES

Here is the simple filter example from L<PerlMx/EXAMPLES> that merely logs
connections.

    package SimpleMinFilter;

    use PerlMx::Simple;
    use base 'PerlMx::Simple';

    sub pmx_connect {
	my $ctx = shift;
	my $from = shift;
	warn "[" . localtime(time) . "] " .
             "SMTP connection from $from\n";
	return CONTINUE;
    }

    1;

Assuming the above is in a file called SimpleMinFilter.pm, here's how to
run the filter (under the C shell):

    % pmx SimpleMinFilter >& SimpleMinFilter.log &

If sendmail has been configured to talk to an inet port other than
3366 (the C<pmx> default port), the port must be specified explicitly:

    % pmx -p inet:12345 SimpleMinFilter >& SimpleMinFilter.log &


=head1 EXPORTS

The following constants are exported by default:

    ACCEPT
    CONTINUE
    DISCARD
    REJECT
    TEMPFAIL

=head1 COPYRIGHT

Copyright (c) 2000-2003 ActiveState Corp.  All rights reserved.

=head1 SEE ALSO

L<PerlMx>

=cut
