package PureMessage::Milter::Ping;

use strict;
use base 'Exporter';
our @EXPORT_OK = qw(pmx_ping);

our $VERSION = '4.000';

use ActiveState::Milter ();
use PureMessage::Version qw(Version);
use Socket qw(pack_sockaddr_un);
use Carp qw(carp);

sub pmx_ping {
    my($port, $timeout, $is_smtp) = @_;
    $timeout ||= 10;

    # parse the port in the same was as libmilter
    my @port;
    if ($port =~ s/^inet://) {
	@port = (($port =~ s/\@(.*)// ? (Host => $1) : ()),
		 Port => $port);
    }
    elsif ($port =~ s/^(unix|local):// || $port =~ m,^/,) {
	@port = (Path => $port);
    }
    elsif ($port =~ /^\d+$/) {
	@port = (Port => $port);
    }
    else {
	carp "Port '$port' not supported";
    }

    my ($status, $msg);

    # connect
    if ($is_smtp) {
        my %args = @port;
	my $sock;
	require IO::Socket;
	if (exists $args{Path}) {
	    $sock = IO::Socket::UNIX->new(Peer => $args{Path},
	                                  Timeout => $timeout);
	}
	else {
	    $args{Host} ||= "localhost";
	    $args{Port} ||= 10025;
	    $sock = IO::Socket::INET->new(PeerAddr => $args{Host},
	                                  PeerPort => $args{Port},
					  Proto    => 'tcp',
	                                  Timeout  => $timeout);
	}
        unless($sock) {
            $@ = $! || 'Unknown connection failure';
            return undef;
        };
	$sock->autoflush;
	my $line;

	do {
	    $line = <$sock>;  # ignore XXX- lines
	} while ($line =~ /^\d\d\d-/);
	$line =~ s/\r?\n\z//;

	unless ($line =~ /^2\d\d\s/) {
	    $@ = "Bad SMTP greeting: '$line'";
	    return undef;
	}

	print $sock "XPMXPING\r\n";

        $line = "";
        $msg = "";
	do {
	    $line = <$sock>;
	    $msg .= $line;
	} while ($line =~ /^\d\d\d-/);

	$msg =~ s/\r?\n\z//;
	unless ($line =~ /^214[ -]/) {
	    $@ = "Bad ping response: '$msg'\nServer may not be PureMessage.";
	    return undef;
	}
	$msg =~ s/(^|\r?\n)214[ -]//g;

	print $sock "QUIT\r\n";
	$sock->close;
    }
    else {
	my $m = ActiveState::Milter->new(@port, Timeout => $timeout)
	    or return undef;

	($status, $msg) = $m->mi_connect("ping.perlmx.activestate.com",
					   pack_sockaddr_un("/PING"));
	unless ($status eq "y") {
	    $@ = "Unexpected status returned from mi_connect '$status'";
	    return undef;
	}

	$msg =~ s/^421 //;  # useless status code
	$msg =~ s/\0\z//;   # NUL-terminated
    }

    return { version => 0 } if $msg eq "PMX";

    unless ($msg =~ /^PMX-[3456789]\./) {
	$@ = "Ping version mismatch: '$msg'";
	return undef;
    }

    # extract information
    my %info;
    @info{qw(version module port stime pid interp flags reinit seq busy extra)}
	= split(' ', $msg, 11);

    if ($info{flags} & 0x04) {
	# these fields is meaningless in fork mode
	delete $info{$_} for qw(interp seq busy);
    }

    if (my $extra = delete $info{extra}) {
	my @extra = split(' ', $extra);
	if ($extra[0] =~ /^v[1-3]\z/) {
	    my $ver = shift(@extra);
	    my %ppool_stats;
	    @ppool_stats{qw(used_slots pool_size total_req max_slots)} = splice(@extra, 0, 4);
	    $ppool_stats{dying_slots} = shift(@extra) if $ver eq "v3";
	    my $total_mcount = 0;
	    while (@extra >= 4) {
		my %slot_stats;
		@slot_stats{qw(pid prev_accepts accepts reqs_pending mcount)} = splice(@extra, 0, 5);
		$slot_stats{retire_count} = shift(@extra) if $ver ne "v1";
		$slot_stats{status} = shift(@extra) if $ver eq "v3";
		delete $slot_stats{pid} unless $slot_stats{pid};
		push(@{$ppool_stats{slots}}, \%slot_stats);
		$total_mcount += $slot_stats{mcount};
	    }
	    $ppool_stats{extra} = \@extra if @extra;
	    $ppool_stats{total_mcount} = $total_mcount;
	    $info{ppool} = \%ppool_stats;
	}
	else {
	    $info{extra} = \@extra;
	}
    }

    $info{version} =~ s/^PMX-//;
    $info{version} = Version($info{version});

    return \%info;
}

1;

=head1 NAME

PureMessage::Milter::Ping - Check status of a running PureMessage milter

=head1 SYNPOSIS

 use PureMessage::Milter::Ping qw(pmx_ping);

 if (my $report = pmx_ping($port)) {
     # the filter is alive
     # and $report is filled with "useful" information
 }
 else {
     # filter is not alive ($@ contains error info)
 }

=head1 DESCRIPTION

Running PureMessage milters can report their state over the milter
protocol.  The C<PureMessage::Milter::Ping> is an interface to this report.  The
module exports a single function:

=over

=item $report = pmx_ping( $port, [ $timeout ] )

The $port argument is that same as for C<pmx>.  It can be specified in
these forms:

  inet:3366@localhost   # connect to port 3366 at localhost
  inet:3366             # same
  3366                  # same
  local:/path           # connect via a Unix domain socket

The timeout value is the number of seconds that we allow the connect
call to take.  If not given it defaults to 10 seconds.

If the milter is not responding, then C<undef> is returned and $@ will
be set to some error string.

If the milter is alive, then a hash reference will be returned.  The
following entries can be filled in:

  version  PureMessage version number
  module   The milter name
  port     What port it listens to
  flags    Current flags
  reinit   Current reinit_limit
  interp   Number of interpreters created
  seq      Number of reinits performed

=back

=head1 BUGS

The 'inet6:' prefix to port is not supported yet.

=head1 COPYRIGHT

Copyright (c) 2002 ActiveState Corp.  All rights reserved.

=head1 SEE ALSO

L<PerlMx>, L<pmx>

=cut
