package PureMessage::Milter::Run;

use strict;
use PureMessage::Version qw($PREFIX);
use base 'Exporter';

our $VERSION = '3.000';

our @EXPORT_OK = qw(start stop restart status running is_running);

my $ETC = "$PREFIX/etc";
our $RUN = "$PREFIX/var/run/milter";

sub mdie {
    my $msg = shift;
    $msg .= "\n" unless $msg =~ /\n\z/;
    (my $me = $0) =~ s,.*/,,;
    die "$me: $msg";
}

sub start {
    my %milter = map { $_ => 0 } @_;

    system("$PREFIX/bin/pmx-config --validate-cache");
    require PureMessage::Config;
    our %config;
    PureMessage::Config->import('%config');
    #use Data::Dump; Data::Dump::dump(\%config);

    my $count_all = 0;
    my $count_enabled = 0;
    my $count_start_attempted = 0;
    my %started;
    my %outfile_size;
    my $failed;

    # Load required modules before forking so that their memory
    # is more likely to be shared if multiple milters are started
    # up at once.
    eval {
	require PerlMx;
	require PureMessage::_sys;
    };
    if ($@) {
	# clean up the message a bit
	$@ =~ s/^Compilation failed in require.*\n\z//m;
	warn $@;
	return 1;
    }

    for my $m (sort keys %{$config{milter}}) {
	if (%milter) {
	    next unless exists $milter{$m};
	    $milter{$m}++;  # mark it
	}
	my $mconf = $config{milter}{$m};
	$count_all++;
	next if !%milter && !$mconf->{enabled};
	$count_enabled++;
	next if is_running($m);

	for my $k (qw(stdout stderr log_to)) {
	    my $f = $mconf->{$k} || next;
	    next if exists $outfile_size{$f};
	    $outfile_size{$f} = -s $f;
	}

	$count_start_attempted++;
	if (my $pid = run($m, $mconf)) {
	    $started{$pid} = $m;
	}
    }

    for my $m (sort keys %milter) {
	next if $milter{$m};
	warn "No configuration for milter '$m'.\n";
    }

    if (%started) {
	# make sure that all started milters run properly
	# before we return

	my $count_started = keys %started;
	require POSIX;
	my $before_time = time;
	my $n = 0;
	my $count_died = 0;
	while ($n++ < (10 * 60) && %started) {
	    sleep(1);

	    # check if any of the children has died
	    my $child;
	    while (($child = waitpid(-1, POSIX::WNOHANG())) > 0) {
		my $m = delete $started{$child} || "with pid $child";
		warn "Milter $m exited with status $?\n";
		$count_died++;
	    }
	    last unless %started;

	    # check if any of the started kids can be pinged
	    for my $pid (keys %started) {
		my $info = is_running($started{$pid}, ping => 1);
		next unless $info && $info->{ping};
		delete $started{$pid};
	    }
	}

	if (%started) {
	    my $t = int(time - $before_time) . "s";
	    for my $m (sort values %started) {
		warn "No response from milter '$m' (waited $t)\n";
	    }
	    $failed++;
	}
	else {
	    $count_started -= $count_died;
	    my $s = $count_started == 1 ? "" : "s";
            my $msg = "$count_started milter$s started";
            if ($count_died) {
                $s = $count_died == 1 ? "" : "s";
                $msg .= ", $count_died milter$s died during startup.\n";
		$msg =~ s/^0 milters started, //;
            }
            else {
                $msg .= ".\n";
            }
	    if ($count_started > 0) {
	        warn $msg;
	    }
	    else {
                warn $msg;
		$failed++;
	    }
	}
    }
    elsif ($count_start_attempted) {
	warn "Failed to start up new milters.\n";
	$failed++;
    }
    elsif (%milter) {
	warn "Requested milters are already running.\n" if $count_all;
    }
    elsif ($count_enabled == 1) {
	warn "The single enabled milter is already running.\n";
    }
    elsif ($count_enabled) {
	warn "The $count_enabled enabled milters are already running.\n";
    }
    elsif ($count_all) {
	warn "No enabled milters in $config{pmx_conf}.\n";
	$failed++;
    }
    else {
	warn "No milters in $config{pmx_conf}.\n";
	$failed++;
    }

  OUTFILE:
    for my $file (sort keys %outfile_size) {
	my $old_size = $outfile_size{$file} || 0;
	my $new_size = (-s $file) || 0;
	next if $new_size <= $old_size;

	my $banner_done = 0;
	if (open(my $f, "<", $file)) {
	    if ($old_size && !seek($f, $old_size, 0)) {
		# seek failed: file probably got rotated, just ignore
		next OUTFILE;
	    }
	    while (<$f>) {
		next if /^=== pmx-mark ===/;
		# skip debug/info/notice messages in log file
		next if /^[012] / and $file =~ /log\z/;
		unless ($banner_done) {
		    ++$banner_done;
		    print "=== $file ===\n";
		}
		print $_;
	    }
	}
    }

    return $failed ? 1 : 0;
}

sub run {
    my($name, $conf) = @_;
    my $pid = fork();
    unless (defined $pid) {
	warn "Can't fork for $name: $!";
	return 0;
    }
    return $pid if $pid;

    # child
    my $filters = $conf->{filter};
    $filters = [$filters] unless ref($filters) eq "ARRAY";
    die "No filters to run.\n" unless @$filters;

    # Try to set up what ps(1) report and initialize syslog
    # XXX: the dynamic linker does strange things on freebsd when $0 is changed
    # XXX: see bug d=40754 for a more detailed explanation
    $0 = "pmx-milter $name" unless ($^O eq 'freebsd');
    PureMessage::_sys::openlog("pmx-milter", "PID", "MAIL");

    # Make sure that additional interpreters that are
    # started pick up the correct configuration settings
    # from PureMessage::Config
    $ENV{PMX_MILTERNAME} = $name;

    # Load filter modules in order to classify them
    my $detector_count = 0;
    for my $f (@$filters) {
	# We test load them in a subprocess so we don't blow
	# up the the main interpreter so much.
	my $pid = open(my $fh, "-|");
	die "Can't fork: $!" unless defined $pid;
	if ($pid) {
	    # parent
	    my $kind = "";
	    while (<$fh>) {
		if (!$kind && /^\#\#([A-Z]+)\#\#$/) {
		    $kind = $1;
		}
		elsif ($_ ne "\n") {
		    print;
		}
	    }
	    close($fh);  # wait for kid

	    if ($kind eq "DETECTOR") {
		$detector_count++;
	    }
	    elsif ($kind eq "OTHER") {
		die "'$f' is not a PureMessage filter module.\n";
	    }
	    elsif ($kind ne "FILTER") {
		# assuming the child has already made some noise, so we
		# don't have to
		print "$kind\n" if $kind;
		exit 1;
	    }
	}
	else {
	    # child
	    eval "#line 3333 \"xyzzy\"\nrequire $f;\n";
	    if ($@) {
		$@ =~ s/(\nCompilation failed in require)? at xyzzy line 3333\.\n?//;
		die "Can't load filter module $f: $@\n";
	    }
	    if ($f->isa('PureMessage::Milter::Scanner::Detector')) {
		print "\n##DETECTOR##\n";
	    }
	    elsif ($f->isa('PerlMx')) {
		print "\n##FILTER##\n";
	    }
	    else {
		print "\n##OTHER##\n";
	    }
	    exit;
	}
    }

    my $filter;
    if ($detector_count) {
	die "Can't mix detectors and real filters yet"
	    if $detector_count != @$filters;
	$filter = "PureMessage::Milter::Scanner";
	# XXX is there a good reason for this?
	die "Can't pass arguments to detectors"
	    if @{$conf->{args}};
	$conf->{args} = $filters;
    }
    else {
	die "Can't run multiple real filters yet"
	    if @$filters > 1;
	$filter = $filters->[0];
    }

    warn "Starting milter $name (pid $$)...\n";

    # off we go...
    require PureMessage::_sys;
    PureMessage::_sys::setsid();
    umask($conf->{umask});
    chdir("$PREFIX/home") or mdie "can't chdir('$PREFIX/home'): $!";

    # redirect standard file descriptors
    open(STDIN, "</dev/null");
    open(STDOUT, ">>", $conf->{stdout})
	or mdie "can't redirect stdout: $!";
    if ($conf->{stdout} eq $conf->{stderr}) {
	open(STDERR, ">&STDOUT");
    }
    else {
	open(STDERR, ">>", $conf->{stderr})
	    or mdie "can't redirect stderr: $!";
    }
    select STDERR; $| = 1;
    select STDOUT; $| = 1;

    # create a pid file
    if (open(my $f, ">", "$RUN/$name.pid")) {
        my $ports = $conf->{port};
	$ports .= " $conf->{smtp_talk_port}" if $conf->{smtp_talk_port};
	print $f "$$ $ports\n";
	unless (close($f)) {
	    warn "Can't write $RUN/$name.pid: $!";
	    unlink("$RUN/$name.pid");  # don't leave half-written file around
	}
    }
    else {
	warn "Can't create $RUN/$name.pid: $!";
    }

    # restore signals trapped by milter core (in case they where set)
    $SIG{$_} = 'DEFAULT' for qw(HUP TERM INT);

    my $flags = 0;
    $flags |= 0x01 if $conf->{clone};
    $flags |= 0x02 if $conf->{concurrency_limit_action} eq 'wait';
    $flags |= 0x04 unless $conf->{use_threads};
    $flags |= 0x08 if $conf->{concurrency_limit_action} eq 'accept';

    my $max_threads = $conf->{concurrency_limit};
    my $preloads = $conf->{preload_interpreters};

    if ($conf->{max_pooled_procs} > 1) {
        $max_threads /= $conf->{max_pooled_procs};
	$max_threads = int($max_threads) + 1
	    if int($max_threads) != $max_threads;  # ceil

	$preloads /= $conf->{min_pooled_procs} || 1;
	$preloads = int($preloads) + 1
	    if int($preloads) != $preloads;  # ceil
    }

    # start the filter
    require PerlMx;
    my $filt = PerlMx::Driver->new($filter,
				   $conf->{port},
				   $max_threads,
				   $flags,
				   [@{$conf->{perl_args}},
				    "--",
				    @{$conf->{args}}
				   ],
				   $preloads,
				   );

    $filt->set_smtp_talk_port($conf->{smtp_talk_port})
        if $conf->{smtp_talk_port};
    $filt->setdbg($conf->{debug_level} - 10)
	if $conf->{debug_level} > 10;
    $filt->setbacklog($conf->{listen_queue_size})
	if $conf->{listen_queue_size};

    $filt->settimeout($conf->{timeout}) if $conf->{timeout};
    $filt->set_max_proto_data_size($conf->{max_proto_data_size})
        if $conf->{max_proto_data_size};
    $filt->reload_count($conf->{reinit_limit});
    $filt->set_thread_stack_size($conf->{thread_stack_size});

    $filt->set_msg_log($conf->{message_log}, $conf->{log_utc})
	if $conf->{message_log};

    if ($conf->{max_pooled_procs}) {
	$filt->proc_pool_opts($conf->{max_pooled_procs},
			      $conf->{min_pooled_procs},
			      $conf->{pooled_proc_idle_limit},
			      $conf->{pooled_proc_connect_limit});
    }

    for my $f (@$filters) {
	my $run_mod = "$f\::Run";
	eval "require $run_mod";
	if ($@) {
	    next if $@ =~ /^Can\'t locate /;
	    die $@;
	}
	$run_mod->milter_start() if $run_mod->can("milter_start");
    }

    my $status = $filt->run;
    unlink("$RUN/$name.pid");

    warn "Milter $name failed: $status\n" if $status;

    for my $f (@$filters) {
	my $run_mod = "$f\::Run";
	$run_mod->milter_stop() if $run_mod->can("milter_stop");
    }

    exit $status;
}

sub is_running {
    my($milter, %opt) = @_;

    open(my $f, "<", "$RUN/$milter.pid")
	|| return undef;
    my $pid_info = <$f> || return undef;
    chomp($pid_info);
    my($milter_pid, $milter_port, $smtp_talk_port) = split(' ', $pid_info);
    return undef unless $milter_pid;

    # is the process still around?
    kill(0, $milter_pid)
	|| return undef;

    # collect extra process information in a hash
    my %info = (pid => $milter_pid);

    if ($opt{ps_info}) {
	require ActiveState::Unix::ProcInfo;
	my $ps = ActiveState::Unix::ProcInfo::proc_info(root_pid =>
							$milter_pid);
	if (my $m = $ps->{$milter_pid}) {
	    $info{vsize} = $m->{vsz};
	    $info{rss} = $m->{rss};
	    $info{threads} = $m->{threads} if $m->{threads};
	    $info{procs} = $m->{descendants};
	}

	for my $pid (keys %$ps) {
	    $info{vsize_procs}{$pid} = $ps->{$pid}{vsz};
	    $info{rss_procs}{$pid} = $ps->{$pid}{rss};
	}
    }
		
    if ($opt{ping} && $milter_port) {
	$info{port} = $milter_port;
	$info{smtp_talk_port} = $smtp_talk_port;
	require PureMessage::Milter::Ping;
	my $report = PureMessage::Milter::Ping::pmx_ping($milter_port,
							 0,
							 $smtp_talk_port);
	if ($report) {
	    $info{ping} = $report;
	}
	else {
	    $info{ping_err} = $@ || '(Unknown Error)';
	}
    }

    return \%info;
}

sub stop {
    my $signal = (@_ && $_[0] =~ s/^-//) ? shift : 1;
    $signal =~ s/^SIG_?//i;
    unless ($signal =~ /^\d+$/) {
	# convert signal name to number
	no strict 'refs';
	require POSIX;
	my $n = eval { &{"POSIX::SIG" . uc($signal)}() };
	die "Unknown signal name '$signal'.\n" unless defined $n;
	$signal = $n;
    }

    # On Linux we need to signal the whole process group, otherwise
    # we will not reach the milter signal catching thread.
    # XXX For soft stops when external processes are running
    # (like external virus scanners) this is not really what we
    # want, as they will also receive the signal (these processes
    # are part of the same process group as the milter).
    $signal = -$signal if $^O eq "linux";

    # Some signal (INT, QUIT, KILL) should affect the whole group.
    $signal = -$signal if $signal == 2 || $signal == 3 || $signal == 9;

    my $killed = 0;
    my $running = running();

    # Check if some of the named milters are not running
    if (@_) {
	for my $m (@_) {
	    next if $running->{$m};
	    warn "Milter '$m' is not running.\n";
	}
    }

    # Collect pids to kill
    my @pids;
    for my $m (keys %$running) {
	next if @_ && !grep $_ eq $m, @_;
	push(@pids, $running->{$m}{pid});
	delete $running->{$m};
    }

    if (@pids) {
	# Do it
	$| = 1;
	print "Sending signal " . abs($signal) .
	    " to process " . ($signal < 0 ? "group " : "") .
	    do { my @p = sort {$a <=> $b} @pids;
		 my $last = pop @p;
		 @p ? join(" and ", join(", ", @p), $last) : $last;
	     } . "...";
	my $killed = kill($signal, @pids);

	# wait for them all to be gone
	my $is_tty = -t;
	my $kids_left;
	for my $n (1 .. 30) {
	    select(undef, undef, undef, 0.5);  # wait
	    $kids_left = running();
	    delete $kids_left->{$_} for keys %$running;
	    last unless %$kids_left;

	    if ($is_tty && $n >= 3) {
		# show ticking dots while we wait
		print ".";
	    }
	}
	print "\n";

	if (!%$kids_left) {
	    my $s = $killed == 1 ? "" : "s";
	    warn "$killed milter$s stopped.\n";
	}
	else {
	    warn "Some milters will not stop!\n";
	    return 1;
	}
    }
    elsif (%$running) {
	warn "None of the named milters are running.\n";
	return 2;
    }
    else {
	warn "No milters are running.\n";
	return 2;
    }

    return 0;
}

sub restart {
    my $hup = "-HUP";
    stop($hup, @_);
    start(@_);
}

sub running {
    opendir(my $d, $RUN) || return;
    my @run_files = readdir($d);
    closedir($d);

    my %running;
    for my $r (@run_files) {
	next unless $r =~ /^(.*)\.pid$/;
	$r = $1;
	if (my $pinfo = is_running($r, @_)) {
	    $running{$r} = $pinfo;
	}
    }
    return \%running;
}

sub status {
    my $show_rss = $ENV{PMX_SHOW_RSS};
    my $running = running(ps_info => 1, ping => 1);
    my $count;
    for my $r (sort keys %$running) {
	$count++;
	my $info = $running->{$r};
	print "\n" if $count > 1;
	print "Milter '$r':\n";
	print "  pid:     $info->{pid}\n";
	print "  port:    $info->{port}\n" if $info->{port};
	print "  smtp_talk_port:    $info->{smtp_talk_port}\n" if $info->{smtp_talk_port};
	if ($info->{vsize}) {
	    printf "  vmsize:  %.1f MB", $info->{vsize} / 1024;
	    printf " (%.1f MB resident)", $info->{rss} / 1024
		if $show_rss && $info->{rss};
	    print "\n";	    

	}
	print "  threads: $info->{threads}\n" if $info->{threads};
	print "  procs:   $info->{procs}\n" if $info->{procs};

	if (my $ping = $info->{ping}) {
	    print  "  interp:  $ping->{interp}\n" if $ping->{interp};
	    print  "  iseq:    $ping->{seq}\n" if $ping->{seq};
	    print  "  busy:    $ping->{busy}\n" if exists $ping->{busy};
	    printf "  flags:   0x%x\n", $ping->{flags};
	    print  "  pingpid: $ping->{pid}\n" if $ping->{pid} && $ping->{pid} != $info->{pid};

	    if (my $stime = $ping->{stime}) {
		require ActiveState::Duration;
		print "  uptime:  ", 
		      ActiveState::Duration::dur_format_eng(time - $stime,
							    0.01, "first"),
                      "\n";
	    }

	    if (my $ppool = $ping->{ppool}) {
		my $dying = "";
		if (my $d = $ppool->{dying_slots}) {
		    $dying = " ($d retiring)";
		}
		print "  ppool:   slots=$ppool->{pool_size}/$ppool->{max_slots}$dying req=$ppool->{total_req}";
		print " msgs=$ppool->{total_mcount}\n";
		my $ix = 0;
		for my $slot (@{$ppool->{slots}}) {
		    $ix++;
		    my $pid = $slot->{pid};
		    my $size = $pid ? $info->{vsize_procs}{$pid} : undef;
		    if ($size) {
			$size = sprintf "%.1f", $size/1024;
			if ($show_rss) {
			    if (my $rss = $info->{rss_procs}{$pid}) {
				$rss = sprintf "%.1f", $rss/1024;
				$size = "$rss/$size";
			    }
			}
		    }
		    my $tot_accepts = $slot->{accepts} + $slot->{prev_accepts};
		    my $procs = "";
		    if (exists $slot->{retire_count}) {
		        $procs = $slot->{retire_count} + 1;
		        $procs = "/$procs";
		    }
		    my $req = $slot->{reqs_pending};
		    if (my $status = $slot->{status}) {
			if ($status == -1) {
			    $req .= "*";
			}
			else {
			    $req .= "+$status";
			}
		    }
		    printf "           %-9s %-10s %-10s %-16s %s\n",
		           "slot#$ix:",
		           ($pid ? "pid=$pid" : "-"),
		           ($size ? "vsz=$size" : "-"),
		           "acc=$slot->{accepts}/$tot_accepts$procs",
		           "req=$req";
		}
	    }

	}
	else {
	    print "  dead?:   $info->{ping_err}\n";
	}
    }
    unless ($count) {
	warn "No milters are running.\n";
	return 2;
    }

    return 0;
}

1;
