package PureMessage::Service::milter;

use strict;
use base 'PureMessage::Service';

use PureMessage::Config qw(%config);
use PureMessage::Service::Config qw( is_stale stale );
use ActiveState::Handy ();
use PureMessage::Milter::Run ();
use PureMessage::Version qw($PREFIX);

# Manual imports so we don't pollute the method name space
*_run = \&ActiveState::Handy::run;
*_running = \&PureMessage::Milter::Run::running;
*_is_running = \&PureMessage::Milter::Run::is_running;

my $pmx_milter_cmd = "$PREFIX/etc/init.d/pmx-milter";

sub Available {
    my $class = shift;

    my @avail;
    my %seen;
    for my $name (sort keys %{$config{milter}}) {
	$seen{$name}++;
	push(@avail, $class->new(name => $name));
    }

    for my $name (sort keys %{_running()}) {
	push(@avail, $class->new(name => $name)) unless $seen{$name};
    }

    return @avail;
}

sub _bg {
    my $pid = fork();
    die "Can't fork: $!" unless defined $pid;
    return $pid if $pid;

    open(STDIN, "</dev/null");
    open(STDOUT, ">/dev/null");
    open(STDERR, ">/dev/null");
    return 0;
}

sub start {
    my($self, %opts) = @_;
    my @cmd = ("$pmx_milter_cmd", "start", $self->name);
    if ($opts{bg}) {
	if (my $pid = _bg()) {
	    return $pid;
	}
	exec(@cmd);
	die "Can't exec: $!";
    }
    _run(@cmd);
}

sub stop {
    my($self, %opts) = @_;
    my @cmd = ("$pmx_milter_cmd", "stop", $self->name);
    if ($opts{bg}) {
	if (my $pid = _bg()) {
	    return $pid;
	}
	exec(@cmd);
	die "Can't exec: $!";
    }
    _run(@cmd);
}

sub restart {
    my($self, %opts) = @_;
    if ($opts{bg}) {
	if (my $pid = _bg()) {
	    return $pid;
	}
	_run("\@-$pmx_milter_cmd", "stop", $self->name);
	exec("$pmx_milter_cmd", "start", $self->name);
	die "Can't exec: $!";
    }
    _run("-$pmx_milter_cmd", "stop", $self->name);
    _run("$pmx_milter_cmd", "start", $self->name);
}

sub status {
    my $self = shift;
    require PureMessage::Milter::Run;
    my $milter = _is_running($self->name, ps_info => 1, ping => 1, @_);
    return { running => 0 } unless $milter;

    my %info = (running => 1);
    $info{Milter} = $milter;
    for (qw(pid port smtp_talk_port)) {
	$info{$_} = delete $milter->{$_};
    }
    if (my $stime = delete $milter->{ping}{stime}) {
	$info{start_time} = $stime;
    }
    if (my $ping = $milter->{ping}) {
	$info{ready} = 1;
	if (my $ppool = $ping->{ppool}) {
	    $info{summary} = "$ppool->{pool_size} procs";
	}
    }
    else {
	$info{ready} = 0;
	$info{summary} = $milter->{ping_err};
    }

    if (my $conf = $config{milter}{$self->name}) {
	$milter->{config} = $conf;
	$info{autostart} = $conf->{enabled};

	require ActiveState::Config::Depend;
	my $mtime = ActiveState::Config::Depend::config_mtime_deps($config{pmx_conf});
	for my $f (@{$conf->{filter}}) {
	    if ($f eq "PureMessage::Milter::Policy") {
		my $m = (stat "$config{prefix}/etc/policy.siv")[9];
		$mtime = $m if $m && $m >= $mtime;
	    }
	}
        my $is_stale_time = is_stale($self->type); # get the highest milter to stale date
        $mtime = $is_stale_time if $is_stale_time and $is_stale_time > $mtime;
	$info{config} = (($info{start_time} || 0) < $mtime) ? "stale" : "sync";
    }
    else {
	$info{autostart} = 0;
	$info{config} = "gone";
    }

    \%info;
}

sub running {
    my $self = shift;
    _is_running($self->name) || 0;
}

1;
